from datetime import datetime, timedelta
import os
from pathlib import Path
from unittest.mock import patch

from chispa import assert_df_equality
from conftest import setup_test_data
from freezegun import freeze_time
import pytest

from pydantic import ValidationError

from pyspark.sql.types import LongType

from koheesio.logger import LoggingFactory
from koheesio.spark.delta import DeltaTableStep, StaleDataCheckStep

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_delta", inherit_from_koheesio=True)


@pytest.mark.parametrize(
    "value,expected",
    [
        (
            "test_catalog.test_schema.test_table",
            {
                "catalog": "test_catalog",
                "database": "test_schema",
                "table": "test_table",
                "table_name": "test_catalog.test_schema.test_table",
            },
        ),
        (
            "test_schema.test_table",
            {
                "catalog": None,
                "database": "test_schema",
                "table": "test_table",
                "table_name": "test_schema.test_table",
            },
        ),
        (
            {"catalog": "a_catalog", "table": "a_schema.a_table"},
            {
                "catalog": "a_catalog",
                "database": "a_schema",
                "table": "a_table",
                "table_name": "a_catalog.a_schema.a_table",
            },
        ),
        (
            "delta_table",
            {
                "catalog": None,
                "database": None,
                "table": "delta_table",
                "table_name": "delta_table",
            },
        ),
    ],
)
def test_table(value, expected):
    if isinstance(value, str):
        actual = DeltaTableStep(table=value)
        log.info(f"actual: {actual}")
    elif isinstance(value, dict):
        actual = DeltaTableStep(**value)
    else:
        raise AssertionError(f"Unable to run test with given input '{value}'")

    assert actual.catalog == expected["catalog"]
    assert actual.database == expected["database"]
    assert actual.table == expected["table"]
    assert actual.table_name == expected["table_name"]

    log.info("delta test completed")


def test_delta_table_properties(spark, setup, delta_file):
    setup_test_data(spark=spark, delta_file=Path(delta_file))
    table_name = "delta_test_table"
    dt = DeltaTableStep(
        table=table_name,
        default_create_properties={
            "delta.randomizeFilePrefixes": "true",
            "delta.checkpoint.writeStatsAsStruct": "true",
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
            "not_string_property_but_bool": True,
            "not_string_property_but_int": 123,
        },
    )

    assert dt.exists is True
    assert dt.dataframe.count() == 10
    assert dt.columns == ["id"]
    assert dt.get_column_type("id") == LongType()
    assert dt.get_column_type("id").simpleString() == "bigint"
    assert dt.get_column_type("none") is None
    assert dt.get_column_type(None) is None
    assert dt.has_change_type is False

    assert dt.table_name == table_name

    properties = dt.get_persisted_properties()
    assert properties.get("delta.enableChangeDataFeed") == "true"
    dt.add_properties({"delta.enableChangeDataFeed": "false"}, override=True)
    assert dt.get_persisted_properties().get("delta.enableChangeDataFeed") == "false"

    dt.add_properties({"delta.enableChangeDataFeed": "true"})
    assert dt.get_persisted_properties().get("delta.enableChangeDataFeed") == "false"

    assert properties.get("test_property") is None
    dt.add_properties({"test_property": "delta_table_value"})
    assert dt.get_persisted_properties().get("test_property") == "delta_table_value"

    assert dt.default_create_properties.get("not_string_property_but_bool") == "true"
    assert dt.default_create_properties.get("not_string_property_but_int") == "123"

    dt.add_properties({"test_property_bool": True})
    assert dt.get_persisted_properties().get("test_property_bool") == "true"
    dt.add_properties({"test_property_int": 123})
    assert dt.get_persisted_properties().get("test_property_int") == "123"

    for v in dt.default_create_properties.values():
        assert isinstance(v, str)


@patch.dict(
    os.environ,
    {
        "DATABRICKS_RUNTIME_VERSION": "lts_11_spark_3_scala_2.12",
    },
)
def test_delta_table_properties_dbx():
    dt = DeltaTableStep(table="delta_test_table_dbx", create_if_not_exists=True)
    assert dt.default_create_properties.get("delta.autoOptimize.autoCompact") == "true"


@pytest.mark.parametrize("value,expected", [("too.many.dots.given.to.be.a.real.table", pytest.raises(ValidationError))])
def test_table_failed(value, expected):
    with expected:
        DeltaTableStep(table=value)

    dt = DeltaTableStep(table="unknown_table")
    assert dt.exists is False


@pytest.mark.parametrize(
    ["table", "create_if_not_exists", "log_level"], [("unknown", False, "DEBUG"), ("unknown", True, "INFO")]
)
def test_exists(caplog, table, create_if_not_exists, log_level):
    with caplog.at_level(log_level):
        dt = DeltaTableStep(table=table, create_if_not_exists=create_if_not_exists)
        dt.log.setLevel(log_level)
        assert dt.exists is False


@pytest.fixture
def test_describe_history_df(spark):
    data = [
        {"version": 0, "timestamp": "2024-12-30 12:00:00", "tableName": "test_table", "operation": "CREATE TABLE"},
        {"version": 1, "timestamp": "2024-12-31 05:29:30", "tableName": "test_table", "operation": "WRITE"},
        {"version": 2, "timestamp": "2025-01-01 11:12:19", "tableName": "test_table", "operation": "MERGE"},
    ]
    return spark.createDataFrame(data)


def test_describe_history__no_limit(mocker, spark, test_describe_history_df):
    mocker.patch.object(DeltaTableStep, "exists", new_callable=mocker.PropertyMock(return_value=True))
    dt = DeltaTableStep(table="test_table")
    mocker.patch.object(spark, "sql", return_value=test_describe_history_df)
    result = dt.describe_history()
    expected_df = spark.createDataFrame(
        [
            {"version": 2, "timestamp": "2025-01-01 11:12:19", "tableName": "test_table", "operation": "MERGE"},
            {"version": 1, "timestamp": "2024-12-31 05:29:30", "tableName": "test_table", "operation": "WRITE"},
            {"version": 0, "timestamp": "2024-12-30 12:00:00", "tableName": "test_table", "operation": "CREATE TABLE"},
        ]
    )
    assert_df_equality(result, expected_df, ignore_column_order=True)


def test_describe_history__with_limit(mocker, spark, test_describe_history_df):
    mocker.patch.object(DeltaTableStep, "exists", new_callable=mocker.PropertyMock(return_value=True))
    dt = DeltaTableStep(table="test_table")
    mocker.patch.object(spark, "sql", return_value=test_describe_history_df)
    result = dt.describe_history(limit=1)
    expected_df = spark.createDataFrame(
        [
            {"version": 2, "timestamp": "2025-01-01 11:12:19", "tableName": "test_table", "operation": "MERGE"},
        ]
    )
    assert_df_equality(result, expected_df, ignore_column_order=True)


def test_describe_history__no_table(mocker):
    mocker.patch.object(DeltaTableStep, "exists", new_callable=mocker.PropertyMock(return_value=False))
    dt = DeltaTableStep(table="test_table")
    result = dt.describe_history()

    assert result is None


@pytest.fixture
def test_stale_data_check_history_df(spark):
    return spark.createDataFrame(
        [
            {
                "timestamp": datetime(year=2024, month=12, day=30, hour=5, minute=29, second=30),
                "operation": "WRITE",
                "version": 1,
            },
            {
                "timestamp": datetime(year=2024, month=12, day=30, hour=5, minute=28, second=30),
                "operation": "CREATE TABLE",
                "version": 0,
            },
        ]
    )


@pytest.mark.parametrize(
    "interval, expected",
    [
        (timedelta(weeks=1), False),  # Not stale, since 1 week > 1 day and some hours ago
        (timedelta(days=2), False),  # Not stale, since 2 days > 1 day and some hours ago
        (timedelta(days=1, hours=7), False),  # Not stale, since 1 day and 7 hours > 1 day, 6 hours, 30 minutes
        (timedelta(days=1, hours=6, minutes=31), False),  # Not stale, since 1 minute more than the timestamp age
        (
            timedelta(days=1, hours=6, minutes=30, seconds=31),
            False,
        ),  # Not stale, since 1 second more than the timestamp age
        (
            timedelta(days=1, hours=6, minutes=30, seconds=30),
            True,
        ),  # Exactly equal to the age, should be considered stale
        (timedelta(days=1, hours=6, minutes=30, seconds=29), True),  # Stale, falls 1 second short
        (timedelta(days=1), True),  # Stale, falls several hours and minutes short
        (
            timedelta(hours=18),
            True,
        ),  # Stale, despite being more than half a day, but less than the full duration since the timestamp
    ],
)
@freeze_time("2024-12-31 12:00:00")
def test_stale_data_check_step__no_refresh_day_num_with_time_components(
    interval, expected, test_stale_data_check_history_df, mocker
):
    mocker.patch("koheesio.spark.delta.DeltaTableStep.describe_history", return_value=test_stale_data_check_history_df)
    assert StaleDataCheckStep(table="dummy_table", interval=interval).execute().is_data_stale == expected


def test_stale_data_check_step__no_table(mocker):
    mocker.patch("koheesio.spark.delta.DeltaTableStep.describe_history", return_value=None)
    assert StaleDataCheckStep(table="dummy_table", interval=timedelta(days=1)).execute().is_data_stale


def test_stale_data_check_step__no_modification_history(mocker, spark):
    history_df = spark.createDataFrame(
        [
            {
                "timestamp": datetime(year=2024, month=12, day=30, hour=5, minute=28, second=30),
                "operation": "CREATE TABLE",
            }
        ]
    )
    mocker.patch("koheesio.spark.delta.DeltaTableStep.describe_history", return_value=history_df)
    assert StaleDataCheckStep(table="dummy_table", interval=timedelta(days=1)).execute().is_data_stale


@pytest.mark.parametrize(
    "interval, refresh_day_num, expected",
    [
        (
            timedelta(days=2, hours=6, minutes=30, seconds=30),
            2,
            False,
        ),  # Data is not stale and it's before the refresh day
        (
            timedelta(days=2, hours=6, minutes=30, seconds=30),
            1,
            True,
        ),  # Data is not stale but it is the refresh day
        (
            timedelta(days=2, hours=6, minutes=30, seconds=30),
            0,
            False,
        ),  # Data is not stale and it is past the refresh day
        (
            timedelta(hours=6, minutes=30, seconds=30),
            2,
            True,
        ),  # Data is stale and it's before the refresh day
        (timedelta(hours=6, minutes=30, seconds=30), 1, True),  # Data is stale and it is the refresh day
        (
            timedelta(hours=6, minutes=30, seconds=30),
            0,
            True,
        ),  # Data is stale and it is past the refresh day
    ],
)
@freeze_time("2024-12-31 12:00:00")  # Tuesday
def test_stale_data_check_step__with_refresh_day(
    interval, refresh_day_num, expected, test_stale_data_check_history_df, mocker
):
    mocker.patch("koheesio.spark.delta.DeltaTableStep.describe_history", return_value=test_stale_data_check_history_df)
    assert (
        StaleDataCheckStep(table="dummy_table", interval=interval, refresh_day_num=refresh_day_num)
        .execute()
        .is_data_stale
        == expected
    )


def test_stale_data_check_step__invalid_refresh_day():
    with pytest.raises(ValueError):
        StaleDataCheckStep(table="dummy_table", interval=timedelta(days=1), refresh_day_num=7).execute()


def test_stale_data_check_step__invalid_staleness_period_with_refresh_day():
    with pytest.raises(ValueError):
        StaleDataCheckStep(table="dummy_table", interval=timedelta(days=10), refresh_day_num=5).execute()



"""
Box Module

The module is used to facilitate various interactions with Box service. The implementation is based on the
functionalities available in Box Python SDK: https://github.com/box/box-python-sdk

Prerequisites
-------------
* Box Application is created in the developer portal using the JWT auth method (Developer Portal - My Apps - Create)
* Application is authorized for the enterprise (Developer Portal - MyApp - Authorization)
"""

from typing import IO, Any, Dict, Optional, Union
from abc import ABC, abstractmethod
from io import BytesIO
from pathlib import PurePath
import re

from boxsdk import Client, JWTAuth
from boxsdk.object.file import File
from boxsdk.object.folder import Folder

from pyspark.sql.functions import expr, lit
from pyspark.sql.types import StructType

from koheesio import Step, StepOutput
from koheesio.models import (
    Field,
    InstanceOf,
    ListOfStrings,
    SecretBytes,
    SecretStr,
    SkipValidation,
    conlist,
    field_validator,
    model_validator,
)
from koheesio.spark.readers import Reader
from koheesio.spark.readers.memory import InMemoryDataReader
from koheesio.spark.writers.buffer import BufferWriter
from koheesio.utils import utc_now


class BoxFolderNotFoundError(Exception):
    """Error when a provided box path does not exist."""


class BoxPathIsEmptyError(Exception):
    """Exception when provided Box path is empty or no files matched the mask."""


class Box(Step, ABC):
    """
    Configuration details required for the authentication can be obtained in the Box Developer Portal by generating
    the Public / Private key pair in "Application Name -> Configuration -> Add and Manage Public Keys".

    The downloaded JSON file will look like this:
    ```json
    {
      "boxAppSettings": {
        "clientID": "client_id",
        "clientSecret": "client_secret",
        "appAuth": {
          "publicKeyID": "public_key_id",
          "privateKey": "private_key",
          "passphrase": "pass_phrase"
        }
      },
      "enterpriseID": "123456"
    }
    ```
    This class is used as a base for the rest of Box integrations, however it can also be used separately to obtain
    the Box client which is created at class initialization.

    Examples
    --------
    ```python
    b = Box(
        client_id="client_id",
        client_secret="client_secret",
        enterprise_id="enterprise_id",
        jwt_key_id="jwt_key_id",
        rsa_private_key_data="rsa_private_key_data",
        rsa_private_key_passphrase="rsa_private_key_passphrase",
    )
    b.client
    ```
    """

    client_id: Union[SecretStr, SecretBytes] = Field(
        default=..., alias="clientID", description="Client ID from the Box Developer console."
    )
    client_secret: Union[SecretStr, SecretBytes] = Field(
        default=..., alias="clientSecret", description="Client Secret from the Box Developer console."
    )
    enterprise_id: Union[SecretStr, SecretBytes] = Field(
        default=..., alias="enterpriseID", description="Enterprise ID from the Box Developer console."
    )
    jwt_key_id: Union[SecretStr, SecretBytes] = Field(
        default=..., alias="publicKeyID", description="PublicKeyID for the public/private generated key pair."
    )
    rsa_private_key_data: Union[SecretStr, SecretBytes] = Field(
        default=..., alias="privateKey", description="Private key generated in the app management console."
    )
    rsa_private_key_passphrase: Union[SecretStr, SecretBytes] = Field(
        default=...,
        alias="passphrase",
        description="Private key passphrase generated in the app management console.",
    )

    client: SkipValidation[Client] = None  # type: ignore

    def init_client(self) -> None:
        """Set up the Box client."""
        if not self.client:
            self.client = Client(JWTAuth(**self.auth_options))

    @property
    def auth_options(self) -> Dict[str, Any]:
        """
        Get a dictionary of authentication options, that can be handily used in the child classes
        """
        return {
            "client_id": self.client_id.get_secret_value(),
            "client_secret": self.client_secret.get_secret_value(),
            "enterprise_id": self.enterprise_id.get_secret_value(),
            "jwt_key_id": self.jwt_key_id.get_secret_value(),
            "rsa_private_key_data": self.rsa_private_key_data.get_secret_value().encode("utf-8"),
            "rsa_private_key_passphrase": self.rsa_private_key_passphrase.get_secret_value(),
        }

    def __init__(self, **data: dict):
        super().__init__(**data)
        self.init_client()

    def execute(self) -> Step.Output:  # type: ignore
        # Plug to be able to unit test ABC
        pass


class BoxFolderBase(Box):
    """
    Generic class to facilitate interactions with Box folders.

    Box SDK provides Folder class that has various properties and methods to interact with Box folders. The object can
    be obtained in multiple ways:
        * provide Box folder identified to `folder` parameter (the identifier can be obtained, for example, from URL)
        * provide existing object to `folder` parameter (boxsdk.object.folder.Folder)
        * provide filesystem-like path to `path` parameter

    See Also
    --------
    boxsdk.object.folder.Folder
    """

    folder: Optional[Union[Folder, str]] = Field(
        default=None, description="Existing folder object or folder identifier"
    )
    path: Optional[str] = Field(default=None, description="Path to the Box folder, for example: `folder/sub-folder/lz")

    root: Optional[Union[Folder, str]] = Field(
        default="0", description="Folder object or identifier of the folder that should be used as root"
    )

    class Output(StepOutput):
        """
        Define outputs for the BoxFolderBase class
        """

        folder: Optional[Folder] = Field(default=None, description="Box folder object")

    @model_validator(mode="after")
    def validate_folder_or_path(self) -> "BoxFolderBase":
        """
        Validations for 'folder' and 'path' parameter usage
        """
        folder_value = self.folder
        path_value = self.path

        if folder_value and path_value:
            raise AttributeError("Cannot user 'folder' and 'path' parameter at the same time")

        if not folder_value and not path_value:
            raise AttributeError("Neither 'folder' nor 'path' parameters are set")

        return self

    @property
    def _obj_from_id(self) -> Folder:
        """
        Get folder object from identifier
        """
        return self.client.folder(folder_id=self.folder).get() if isinstance(self.folder, str) else self.folder

    def action(self) -> Optional[Folder]:
        """
        Placeholder for 'action' method, that should be implemented in the child classes

        Returns
        -------
            Folder or None
        """
        raise NotImplementedError

    def execute(self) -> Output:
        self.output.folder = self.action()


class BoxFolderGet(BoxFolderBase):
    """
    Get the Box folder object for an existing folder or create a new folder and parent directories.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxFolderGet

    auth_params = {...}
    folder = BoxFolderGet(**auth_params, path="/foo/bar").execute().folder
    # or
    folder = BoxFolderGet(**auth_params, path="1").execute().folder
    ```
    """

    create_sub_folders: Optional[bool] = Field(
        False, description="Create sub-folders recursively if the path does not exist."
    )

    def _get_or_create_folder(self, current_folder_object: Folder, next_folder_name: str) -> Folder:
        """
        Get or create a folder.

        Parameters
        ----------
        current_folder_object: Folder
            Current folder object.
        next_folder_name: str
            Name of the next folder.

        Returns
        -------
        next_folder_object: Folder
            Next folder object.

        Raises
        ------
        BoxFolderNotFoundError
            If the folder does not exist and 'create_sub_folders' is set to False.
        """
        for item in current_folder_object.get_items():
            # noinspection PyUnresolvedReferences
            if item.type == "folder" and item.name == next_folder_name:
                # noinspection PyTypeChecker
                return item

        if self.create_sub_folders:
            return current_folder_object.create_subfolder(name=next_folder_name)
        else:
            raise BoxFolderNotFoundError(
                f"The folder '{next_folder_name}' does not exist! Set 'create_sub_folders' to True "
                "to create required directory structure automatically."
            )

    def action(self) -> Optional[Folder]:
        """
        Get folder action

        Returns
        -------
        folder: Optional[Folder]
            Box Folder object as specified in Box SDK
        """
        current_folder_object = None

        if self.folder:
            current_folder_object = self._obj_from_id

        if self.path:
            cleaned_path_parts = [p for p in PurePath(self.path).parts if p.strip() not in [None, "", " ", "/"]]
            current_folder_object: Union[Folder, str] = (
                self.client.folder(folder_id=self.root) if isinstance(self.root, str) else self.root
            )

            for next_folder_name in cleaned_path_parts:
                current_folder_object = self._get_or_create_folder(current_folder_object, next_folder_name)

        self.log.info(f"Folder identified or created: {current_folder_object}")
        return current_folder_object


class BoxFolderCreate(BoxFolderGet):
    """
    Explicitly create the new Box folder object and parent directories.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxFolderCreate

    auth_params = {...}
    folder = BoxFolderCreate(**auth_params, path="/foo/bar").execute()
    ```
    """

    create_sub_folders: bool = Field(
        default=True, description="Create sub-folders recursively if the path does not exist."
    )

    @field_validator("folder")
    def validate_folder(cls, folder: Any) -> None:
        """
        Validate 'folder' parameter
        """
        if folder:
            raise AttributeError("Only 'path' parameter is allowed in the context of folder creation.")


class BoxFolderDelete(BoxFolderBase):
    """
    Delete existing Box folder based on object, identifier or path.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxFolderDelete

    auth_params = {...}
    BoxFolderDelete(**auth_params, path="/foo/bar").execute()
    # or
    BoxFolderDelete(**auth_params, folder="1").execute()
    # or
    folder = BoxFolderGet(**auth_params, path="/foo/bar").execute().folder
    BoxFolderDelete(**auth_params, folder=folder).execute()
    ```
    """

    def action(self) -> None:
        """
        Delete folder action

        Returns
        -------
            None
        """
        if self.folder:
            folder = self._obj_from_id
        else:  # path
            folder = BoxFolderGet.from_step(self).action()

        self.log.info(f"Deleting Box folder '{folder}'...")
        folder.delete()


class BoxReaderBase(Box, Reader, ABC):
    """
    Base class for Box readers.
    """

    schema_: Optional[StructType] = Field(
        default=None,
        alias="schema",
        description="[Optional] Schema that will be applied during the creation of Spark DataFrame",
    )
    params: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="[Optional] Set of extra parameters that should be passed to the Spark reader.",
    )

    file_encoding: Optional[str] = Field(
        default="utf-8", description="[Optional] Set file encoding format. By default is utf-8."
    )


class BoxCsvFileReader(BoxReaderBase):
    """
    Class facilitates reading one or multiple CSV files with the same structure directly from Box and
    producing Spark Dataframe.

    Notes
    -----
    To manually identify the ID of the file in Box, open the file through Web UI, and copy ID from
    the page URL, e.g. https://foo.ent.box.com/file/1234567890 , where 1234567890 is the ID.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxCsvFileReader
    from pyspark.sql.types import StructType

    schema = StructType(...)
    b = BoxCsvFileReader(
        client_id="",
        client_secret="",
        enterprise_id="",
        jwt_key_id="",
        rsa_private_key_data="",
        rsa_private_key_passphrase="",
        file=["1", "2"],
        schema=schema,
    ).execute()
    b.df.show()
    ```
    """

    files: ListOfStrings = Field(default=..., alias="file", description="ID or list of IDs for the files to read.")

    def execute(self) -> BoxReaderBase.Output:
        """
        Loop through the list of provided file identifiers and load data into dataframe.
        For traceability purposes the following columns will be added to the dataframe:
            * meta_file_id: the identifier of the file on Box
            * meta_file_name: name of the file

        Returns
        -------
        DataFrame
        """
        df = None
        for f in self.files:
            self.log.debug(f"Currently processing file with ID '{f}'")
            file = self.client.file(file_id=f)

            self.log.debug(f"Reading file with ID '{file.object_id}' and name '{file.get().name}' into Spark DataFrame")
            data = file.content().decode(self.file_encoding)
            temp_df = InMemoryDataReader(data=data, schema=self.schema_, params=self.params, format="csv").read()

            # type: ignore
            # noinspection PyUnresolvedReferences
            temp_df = (
                temp_df
                # fmt: off
                .withColumn("meta_file_id", lit(file.object_id))
                .withColumn("meta_file_name", lit(file.get().name))
                .withColumn("meta_load_timestamp", expr("to_utc_timestamp(current_timestamp(), current_timezone())"))
                # fmt: on
            )

            df = temp_df if not df else df.union(temp_df)

        self.output.df = df


class BoxCsvPathReader(BoxReaderBase):
    """
    Read all CSV files from the specified path into the dataframe. Files can be filtered using the regular expression
    in the 'filter' parameter. The default behavior is to read all CSV / TXT files from the specified path.

    Notes
    -----
    The class does not contain archival capability as it is presumed that the user wants to make sure that the full
    pipeline is successful (for example, the source data was transformed and saved) prior to moving the source
    files. Use BoxToBoxFileMove class instead and provide the list of IDs from 'file_id' output.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxCsvPathReader

    auth_params = {...}
    b = BoxCsvPathReader(**auth_params, path="foo/bar/").execute()
    b.df  # Spark Dataframe
    ...  # do something with the dataframe
    from koheesio.steps.integrations.box import BoxToBoxFileMove

    bm = BoxToBoxFileMove(
        **auth_params, file=b.file_id, path="/foo/bar/archive"
    )
    ```
    """

    path: str = Field(default=..., description="Box path")
    filter: str = Field(default=r".csv|.txt$", description="[Optional] Regexp to filter folder contents")

    def execute(self) -> BoxReaderBase.Output:
        """
        Identify the list of files from the source Box path that match desired filter and load them into Dataframe
        """
        folder = BoxFolderGet.from_step(self).execute().folder

        # Identify the list of files that should be processed
        files = [item for item in folder.get_items() if item.type == "file" and re.search(self.filter, item.name)]

        if len(files) > 0:
            self.log.info(
                f"A total of {len(files)} files, that match the mask '{self.filter}' has been detected in {self.path}."
                f" They will be loaded into Spark Dataframe: {files}"
            )
        else:
            raise BoxPathIsEmptyError(f"Path '{self.path}' is empty or none of files match the mask '{self.filter}'")

        file = [file_id.object_id for file_id in files]
        self.output.df = BoxCsvFileReader.from_step(self, file=file).read()
        self.output.file = file  # e.g. if files should be archived after pipeline is successful


class BoxFileBase(Box):
    """
    Generic class to facilitate interactions with Box folders.

    Box SDK provides File class that has various properties and methods to interact with Box files. The object can
    be obtained in multiple ways:
        * provide Box file identified to `file` parameter (the identifier can be obtained, for example, from URL)
        * provide existing object to `file` parameter (boxsdk.object.file.File)

    Notes
    -----
    Refer to BoxFolderBase for mor info about `folder` and `path` parameters

    See Also
    --------
    boxsdk.object.file.File
    """

    files: conlist(Union[File, str], min_length=1) = Field(
        default=..., alias="file", description="List of Box file objects or identifiers"
    )

    folder: Optional[Union[Folder, str]] = Field(
        default=None, description="Existing folder object or folder identifier"
    )
    path: Optional[str] = Field(default=None, description="Path to the Box folder, for example: `folder/sub-folder/lz")

    def action(self, file: File, folder: Folder) -> None:
        """
        Abstract class for File level actions.
        """
        raise NotImplementedError

    def execute(self) -> Box.Output:
        """
        Generic execute method for all BoxToBox interactions. Deals with getting the correct folder and file objects
        from various parameter inputs
        """
        if self.path:
            _folder = BoxFolderGet.from_step(self).execute().folder
        else:
            _folder = self.client.folder(folder_id=self.folder) if isinstance(self.folder, str) else self.folder

        for _file in self.files:
            _file = self.client.file(file_id=_file) if isinstance(_file, str) else _file
            self.action(file=_file, folder=_folder)


class BoxToBoxFileCopy(BoxFileBase):
    """
    Copy one or multiple files to the target Box path.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxToBoxFileCopy

    auth_params = {...}
    BoxToBoxFileCopy(
        **auth_params, file=["1", "2"], path="/foo/bar"
    ).execute()
    # or
    BoxToBoxFileCopy(**auth_params, file=["1", "2"], folder="1").execute()
    # or
    folder = BoxFolderGet(**auth_params, path="/foo/bar").execute().folder
    BoxToBoxFileCopy(
        **auth_params, file=[File(), File()], folder=folder
    ).execute()
    ```
    """

    def action(self, file: File, folder: Folder) -> None:
        """
        Copy file to the desired destination and extend file description with the processing info

        Parameters
        ----------
        file : File
            File object as specified in Box SDK
        folder : Folder
            Folder object as specified in Box SDK
        """
        self.log.info(f"Copying '{file.get()}' to '{folder.get()}'...")
        file.copy(parent_folder=folder).update_info(
            data={"description": "\n".join([f"File processed on {utc_now()}", file.get()["description"]])}
        )


class BoxToBoxFileMove(BoxFileBase):
    """
    Move one or multiple files to the target Box path

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxToBoxFileMove

    auth_params = {...}
    BoxToBoxFileMove(
        **auth_params, file=["1", "2"], path="/foo/bar"
    ).execute()
    # or
    BoxToBoxFileMove(**auth_params, file=["1", "2"], folder="1").execute()
    # or
    folder = BoxFolderGet(**auth_params, path="/foo/bar").execute().folder
    BoxToBoxFileMove(
        **auth_params, file=[File(), File()], folder=folder
    ).execute()
    ```
    """

    def action(self, file: File, folder: Folder) -> None:
        """
        Move file to the desired destination and extend file description with the processing info

        Parameters
        ----------
        file : File
            File object as specified in Box SDK
        folder : Folder
            Folder object as specified in Box SDK
        """
        self.log.info(f"Moving '{file.get()}' to '{folder.get()}'...")
        file.move(parent_folder=folder).update_info(
            data={"description": "\n".join([f"File processed on {utc_now()}", file.get()["description"]])}
        )


class BoxBaseFileWriter(BoxFolderBase, ABC):
    """
    Base class for writing files to Box
    """

    overwrite: bool = Field(default=False, description="Overwrite the file if it exists on box")
    description: Optional[str] = Field(None, description="Optional description to add to the file in Box")

    class Output(StepOutput):
        """Base Output class for Koheesio Box File Writers."""

        file: File = Field(default=..., description="File object in Box")
        shared_link: str = Field(default=..., description="Shared link for the Box file")

    @abstractmethod
    def action(self):
        raise NotImplementedError(
            "You are not supposed to use this class directly, this serves as a base model for Koheesio Box file writers"
        )

    def execute(self) -> Output:
        self.action()

    def get_file(self, folder: Folder, file_name: str) -> Union[File, None]:
        """
        Get the file object, with ID, if the file exisst
        else return none
        """
        self.log.info(f"We are looking for {file_name}")

        items = folder.get_items()
        for item in items:
            if item.type == "file" and item.name == file_name:
                self.log.info("File with the same name is found")
                return item
        return None

    def write_or_overwrite_file_with_stream(self, folder: Folder, file_stream: IO[bytes], file_name: str) -> File:
        """
        Overwrite the file if it exists, else write a new file
        Returns the new file object
        """
        curr_file = self.get_file(folder=folder, file_name=file_name)
        if curr_file and self.overwrite:
            self.log.info("Overwriting to box ...")
            new_file = curr_file.update_contents_with_stream(file_stream=file_stream)
        else:
            self.log.info("Writing to box ...")
            folder.preflight_check(size=0, name=file_name)
            new_file = folder.upload_stream(
                file_stream=file_stream, file_name=file_name, file_description=self.description
            )

        return new_file

    def write_file(self, file_stream: IO[bytes], file_name: str) -> File:
        folder: Folder = BoxFolderGet.from_step(self, create_sub_folders=True).execute().folder

        # noinspection PyUnresolvedReferences
        self.log.info(f"Uploading file '{file_name}' to Box folder '{folder.get().name}'...")
        _box_file: File = self.write_or_overwrite_file_with_stream(
            folder=folder, file_stream=file_stream, file_name=file_name
        )

        self.output.file = _box_file
        self.output.shared_link = _box_file.get_shared_link()


class BoxFileWriter(BoxBaseFileWriter):
    """
    Write file or a file-like object to Box.

    Examples
    --------
    ```python
    from koheesio.steps.integrations.box import BoxFileWriter

    auth_params = {...}
    f1 = BoxFileWriter(
        **auth_params, path="/foo/bar", file="path/to/my/file.ext"
    ).execute()
    # or
    import io

    b = io.BytesIO(b"my-sample-data")
    f2 = BoxFileWriter(
        **auth_params, path="/foo/bar", file=b, name="file.ext"
    ).execute()
    ```
    """

    file: Union[str, BytesIO] = Field(default=..., description="Path to file or a file-like object")
    file_name: Optional[str] = Field(
        default=None,
        description="When file path or name is provided to 'file' parameter, this will override the original name."
        "When binary stream is provided, the 'name' should be used to set the desired name for the Box file.",
    )

    @model_validator(mode="before")
    def validate_name_for_binary_data(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate 'file_name' parameter when providing a binary input for 'file'."""
        file, file_name = values.get("file"), values.get("file_name")
        if not isinstance(file, str) and not file_name:
            raise AttributeError("The parameter 'file_name' is mandatory when providing a binary input for 'file'.")

        return values

    def action(self) -> None:
        _file = self.file
        _name = self.file_name

        if isinstance(_file, str):
            _name = _name if _name else PurePath(_file).name
            with open(_file, "rb") as f:
                _file = BytesIO(f.read())

        self.write_file(file_stream=_file, file_name=_name)


class BoxBufferFileWriter(BoxBaseFileWriter):
    """
    Buffer version of the BoxFileWriter. This class relies on BufferWriter implementations class to write
    the file content into a buffer first, and then it will reuse this to source data to be written to box.

    Examples
    --------
    ```python
    from koheesio.spark.writers.buffer import PandasCsvBufferWriter

    df = spark.createDataFrame([("A", 1), ("B", 2)], ["foo", "bar"])
    buffer_writer = PandasCsvBufferWriter(df=df)

    auth_params = {...}

    box_buffer_writer = BoxBufferFileWriter(
        **auth_params,
        path="/foo",
        file_name="upload.txt",
        buffer_writer=buffer_writer,
    )

    f = box_buffer_writer.execute()
    ```
    """

    buffer_writer: InstanceOf[BufferWriter] = Field(
        default=..., description="Koheesio buffer writer that will be used to produce the output"
    )
    file_name: str = Field(default=..., description="Name to be used for the Box file")

    def action(self):
        # Writes the data to the buffer using the provided buffer writer
        self.buffer_writer.write()
        self.buffer_writer.output.rewind_buffer()

        self.write_file(file_stream=self.buffer_writer.output.buffer, file_name=self.file_name)



# noinspection PyUnresolvedReferences
"""
Snowflake steps and tasks for Koheesio

Every class in this module is a subclass of `Step` or `BaseModel` and is used to perform operations on Snowflake.

Notes
-----
Every Step in this module is based on [SnowflakeBaseModel](./snowflake.md#koheesio.spark.snowflake.SnowflakeBaseModel).
The following parameters are available for every Step.

Parameters
----------
url : str
    Hostname for the Snowflake account, e.g. <account>.snowflakecomputing.com.
    Alias for `sfURL`.
user : str
    Login name for the Snowflake user.
    Alias for `sfUser`.
password : SecretStr
    Password for the Snowflake user.
    Alias for `sfPassword`.
database : str
    The database to use for the session after connecting.
    Alias for `sfDatabase`.
sfSchema : str
    The schema to use for the session after connecting.
    Alias for `schema` ("schema" is a reserved name in Pydantic, so we use `sfSchema` as main name instead).
role : str
    The default security role to use for the session after connecting.
    Alias for `sfRole`.
warehouse : str
    The default virtual warehouse to use for the session after connecting.
    Alias for `sfWarehouse`.
authenticator : Optional[str], optional, default=None
    Authenticator for the Snowflake user. Example: "okta.com".
options : Optional[Dict[str, Any]], optional, default={"sfCompress": "on", "continue_on_error": "off"}
    Extra options to pass to the Snowflake connector.
format : str, optional, default="snowflake"
    The default `snowflake` format can be used natively in Databricks, use `net.snowflake.spark.snowflake` in other
    environments and make sure to install required JARs.
"""

from __future__ import annotations

from typing import Any, Dict, Generator, List, Optional, Set, Union
from abc import ABC
from contextlib import contextmanager
from functools import partial
import os
import tempfile
from types import ModuleType
from urllib.parse import urlparse

from koheesio import Step
from koheesio.logger import warn
from koheesio.models import (
    BaseModel,
    ExtraParamsMixin,
    Field,
    PrivateAttr,
    SecretStr,
    conlist,
    field_validator,
    model_validator,
)
from koheesio.spark.utils.common import on_databricks

__all__ = [
    "GrantPrivilegesOnFullyQualifiedObject",
    "GrantPrivilegesOnObject",
    "GrantPrivilegesOnTable",
    "GrantPrivilegesOnView",
    "SnowflakeRunQueryPython",
    "SnowflakeBaseModel",
    "SnowflakeStep",
    "SnowflakeTableStep",
    "safe_import_snowflake_connector",
]

# pylint: disable=inconsistent-mro, too-many-lines
# Turning off inconsistent-mro because we are using ABCs and Pydantic models and Tasks together in the same class
# Turning off too-many-lines because we are defining a lot of classes in this file


def __check_access_snowflake_config_dir() -> bool:
    """Check if the Snowflake configuration directory is accessible

    Returns
    -------
    bool
        True if the Snowflake configuration directory is accessible, otherwise False

    Raises
    ------
    RuntimeError
        If `snowflake-connector-python` is not installed
    """
    check_result = False

    try:
        from snowflake.connector.sf_dirs import _resolve_platform_dirs  # noqa: F401

        _resolve_platform_dirs().user_config_path
        check_result = True
    except PermissionError as e:
        warn(f"Snowflake configuration directory is not accessible. Please check the permissions.Catched error: {e}")
    except (ImportError, ModuleNotFoundError) as e:
        raise RuntimeError(
            "You need to have the `snowflake-connector-python` package installed to use the Snowflake steps that are"
            "based around SnowflakeRunQueryPython. You can install this in Koheesio by adding `koheesio[snowflake]` to "
            "your package dependencies.",
        ) from e

    return check_result


def safe_import_snowflake_connector() -> Optional[ModuleType]:
    """Validate that the Snowflake connector is installed

    Returns
    -------
    Optional[ModuleType]
        The Snowflake connector module if it is installed, otherwise None
    """
    is_accessable_sf_conf_dir = __check_access_snowflake_config_dir()

    if not is_accessable_sf_conf_dir and on_databricks():
        snowflake_home: str = tempfile.mkdtemp(prefix="snowflake_tmp_", dir="/tmp")  # nosec B108:ignore bandit check for CWE-377
        os.environ["SNOWFLAKE_HOME"] = snowflake_home
        warn(f"Getting error for snowflake config directory. Going to use temp directory `{snowflake_home}` instead.")
    elif not is_accessable_sf_conf_dir:
        raise PermissionError("Snowflake configuration directory is not accessible. Please check the permissions.")

    try:
        # Keep the import here as it is perfroming resolution of snowflake configuration directory
        from snowflake import connector as snowflake_connector

        return snowflake_connector
    except (ImportError, ModuleNotFoundError):
        warn(
            "You need to have the `snowflake-connector-python` package installed to use the Snowflake steps that are"
            "based around SnowflakeRunQueryPython. You can install this in Koheesio by adding `koheesio[snowflake]` to "
            "your package dependencies.",
            UserWarning,
        )
        return None


SF_DEFAULT_PARAMS = {"sfCompress": "on", "continue_on_error": "off"}


class SnowflakeBaseModel(BaseModel, ExtraParamsMixin, ABC):  # type: ignore[misc]
    """
    BaseModel for setting up Snowflake Driver options.

    Notes
    -----
    * Snowflake is supported natively in Databricks 4.2 and newer:
        https://docs.snowflake.com/en/user-guide/spark-connector-databricks
    * Refer to Snowflake docs for the installation instructions for non-Databricks environments:
        https://docs.snowflake.com/en/user-guide/spark-connector-install
    * Refer to Snowflake docs for connection options:
        https://docs.snowflake.com/en/user-guide/spark-connector-use#setting-configuration-options-for-the-connector

    Parameters
    ----------
    url : str
        Hostname for the Snowflake account, e.g. <account>.snowflakecomputing.com.
        Alias for `sfURL`.
    user : str
        Login name for the Snowflake user.
        Alias for `sfUser`.
    password : SecretStr
        Password for the Snowflake user.
        Alias for `sfPassword`.
    role : str
        The default security role to use for the session after connecting.
        Alias for `sfRole`.
    warehouse : str
        The default virtual warehouse to use for the session after connecting.
        Alias for `sfWarehouse`.
    authenticator : Optional[str], optional, default=None
        Authenticator for the Snowflake user. Example: "okta.com".
    database : Optional[str], optional, default=None
        The database to use for the session after connecting.
        Alias for `sfDatabase`.
    sfSchema : Optional[str], optional, default=None
        The schema to use for the session after connecting.
        Alias for `schema` ("schema" is a reserved name in Pydantic, so we use `sfSchema` as main name instead).
    options : Optional[Dict[str, Any]], optional, default={"sfCompress": "on", "continue_on_error": "off"}
        Extra options to pass to the Snowflake connector.
    """

    url: str = Field(
        default=...,
        alias="sfURL",
        description="Hostname for the Snowflake account, e.g. <account>.snowflakecomputing.com",
        examples=["example.snowflakecomputing.com"],
    )
    user: str = Field(default=..., alias="sfUser", description="Login name for the Snowflake user")
    password: SecretStr = Field(default=..., alias="sfPassword", description="Password for the Snowflake user")
    role: str = Field(
        default=..., alias="sfRole", description="The default security role to use for the session after connecting"
    )
    warehouse: str = Field(
        default=...,
        alias="sfWarehouse",
        description="The default virtual warehouse to use for the session after connecting",
    )
    authenticator: Optional[str] = Field(
        default=None,
        description="Authenticator for the Snowflake user",
        examples=["okta.com"],
    )
    database: Optional[str] = Field(
        default=None, alias="sfDatabase", description="The database to use for the session after connecting"
    )
    sfSchema: Optional[str] = Field(
        default=..., alias="schema", description="The schema to use for the session after connecting"
    )
    params: Optional[Dict[str, Any]] = Field(
        default_factory=partial(dict, **SF_DEFAULT_PARAMS),
        description="Extra options to pass to the Snowflake connector, by default it includes "
        "'sfCompress': 'on' and  'continue_on_error': 'off'",
        alias="options",
        examples=[{"sfCompress": "on", "continue_on_error": "off"}],
    )

    @property
    def options(self) -> Dict[str, Any]:
        """Shorthand for accessing self.params provided for backwards compatibility"""
        return self

    def get_options(self, by_alias: bool = True, include: Optional[Set[str]] = None) -> Dict[str, Any]:
        """Get the sfOptions as a dictionary.

        Note
        ----
        - Any parameters that are `None` are excluded from the output dictionary.
        - `sfSchema` and `password` are handled separately.
        - The values from both 'options' and 'params' (kwargs / extra params) are included as is.
        - Koheesio specific fields are excluded by default (i.e. `name`, `description`, `format`).

        Parameters
        ----------
        by_alias : bool, optional, default=True
            Whether to use the alias names or not. E.g. `sfURL` instead of `url`
        include : Optional[Set[str]], optional, default=None
            Set of keys to include in the output dictionary. When None is provided, all fields will be returned.
            Note: be sure to include all the keys you need.
        """
        exclude_set = {
            # Exclude koheesio specific fields
            "name",
            "description",
            # params are separately implemented
            "params",
            "options",
            # schema and password have to be handled separately
            "sfSchema",
            "password",
        } - (include or set())

        fields = self.model_dump(
            by_alias=by_alias,
            exclude_none=True,
            exclude=exclude_set,
        )

        # handle schema and password
        fields.update(
            {
                "sfSchema" if by_alias else "schema": self.sfSchema,
                "sfPassword" if by_alias else "password": self.password.get_secret_value(),
            }
        )

        # handle include
        if include:
            # user specified filter
            fields = {key: value for key, value in fields.items() if key in include}
        else:
            # default filter
            include = {"params"}

        # handle options
        if "options" in include:
            options = fields.pop("params", self.params)
            fields.update(**options)

        # handle params
        if "params" in include:
            params = fields.pop("params", self.params)
            fields.update(**params)

        return {key: value for key, value in fields.items() if value}


class SnowflakeStep(SnowflakeBaseModel, Step, ABC):
    """Expands the SnowflakeBaseModel so that it can be used as a Step"""


class SnowflakeTableStep(SnowflakeStep, ABC):
    """Expands the SnowflakeStep, adding a 'table' parameter"""

    table: str = Field(default=..., description="The name of the table")

    @property
    def full_name(self) -> str:
        """
        Returns the fullname of snowflake table based on schema and database parameters.

        Returns
        -------
        str
            Snowflake Complete table name (database.schema.table)
        """
        return f"{self.database}.{self.sfSchema}.{self.table}"


class SnowflakeRunQueryPython(SnowflakeStep):
    """
    Run a query on Snowflake using the Python connector

    Example
    -------
    ```python
    RunQueryPython(
        database="MY_DB",
        schema="MY_SCHEMA",
        warehouse="MY_WH",
        user="account",
        password="***",
        role="APPLICATION.SNOWFLAKE.ADMIN",
        query="CREATE TABLE test (col1 string)",
    ).execute()
    ```
    """

    query: str = Field(default=..., description="The query to run", alias="sql", serialization_alias="query")
    account: Optional[str] = Field(default=None, description="Snowflake Account Name", alias="account")

    # for internal use
    _snowflake_connector: Optional[ModuleType] = PrivateAttr(default_factory=safe_import_snowflake_connector)

    class Output(SnowflakeStep.Output):
        """Output class for RunQueryPython"""

        results: List = Field(default_factory=list, description="The results of the query")

    @model_validator(mode="before")
    def _validate_account(cls, values: Dict) -> Dict:
        """Populate account from URL if not provided"""
        if not values.get("account"):
            parsed_url = urlparse(values.get("url") or values.get("sfURL"))
            base_url = parsed_url.hostname or parsed_url.path
            values["account"] = base_url.split(".")[0]

        return values

    @field_validator("query")
    def validate_query(cls, query: str) -> str:
        """Replace escape characters, strip whitespace, ensure it is not empty"""
        query = query.replace("\\n", "\n").replace("\\t", "\t").strip()
        if not query:
            raise ValueError("Query cannot be empty")
        return query

    def get_options(self, by_alias: bool = False, include: Optional[Set[str]] = None) -> Dict[str, Any]:
        if include is None:
            include = {
                "account",
                "url",
                "authenticator",
                "user",
                "role",
                "warehouse",
                "database",
                "schema",
                "password",
            }
        return super().get_options(by_alias=by_alias, include=include)

    @property
    @contextmanager
    def conn(self) -> Generator:
        if not self._snowflake_connector:
            raise RuntimeError("Snowflake connector is not installed. Please install `snowflake-connector-python`.")

        sf_options = self.get_options()
        _conn = self._snowflake_connector.connect(**sf_options)
        self.log.info(f"Connected to Snowflake account: {sf_options['account']}")

        try:
            from snowflake.connector.connection import logger as snowflake_logger

            _preserve_snowflake_logger = snowflake_logger
            snowflake_logger = self.log
            snowflake_logger.debug("Replace snowflake logger with Koheesio logger")
            yield _conn
        finally:
            if _preserve_snowflake_logger:
                if snowflake_logger:
                    snowflake_logger.debug("Restore snowflake logger")
                snowflake_logger = _preserve_snowflake_logger

            if _conn:
                _conn.close()

    def get_query(self) -> str:
        """allows to customize the query"""
        return self.query

    def execute(self) -> None:
        """Execute the query"""
        with self.conn as conn:
            cursors = conn.execute_string(
                self.get_query(),
            )
            for cursor in cursors:
                self.log.debug(f"Cursor executed: {cursor}")
                self.output.results.extend(cursor.fetchall())


class GrantPrivilegesOnObject(SnowflakeRunQueryPython):
    """
    A wrapper on Snowflake GRANT privileges

    With this Step, you can grant Snowflake privileges to a set of roles on a table, a view, or an object

    See Also
    --------
    https://docs.snowflake.com/en/sql-reference/sql/grant-privilege.html

    Parameters
    ----------
    account : str
        Snowflake Account Name.
    warehouse : str
        The name of the warehouse. Alias for `sfWarehouse`
    user : str
        The username. Alias for `sfUser`
    password : SecretStr
        The password. Alias for `sfPassword`
    role : str
        The role name
    object : str
        The name of the object to grant privileges on
    type : str
        The type of object to grant privileges on, e.g. TABLE, VIEW
    privileges : Union[conlist(str, min_length=1), str]
        The Privilege/Permission or list of Privileges/Permissions to grant on the given object.
    roles : Union[conlist(str, min_length=1), str]
        The Role or list of Roles to grant the privileges to

    Example
    -------
    ```python
    GrantPermissionsOnTable(
        object="MY_TABLE",
        type="TABLE",
        warehouse="MY_WH",
        user="gid.account@abc.com",
        password=Secret("super-secret-password"),
        role="APPLICATION.SNOWFLAKE.ADMIN",
        permissions=["SELECT", "INSERT"],
    ).execute()
    ```

    In this example, the `APPLICATION.SNOWFLAKE.ADMIN` role will be granted `SELECT` and `INSERT` privileges on
    the `MY_TABLE` table using the `MY_WH` warehouse.
    """

    object: str = Field(default=..., description="The name of the object to grant privileges on")
    type: str = Field(default=..., description="The type of object to grant privileges on, e.g. TABLE, VIEW")

    privileges: Union[conlist(str, min_length=1), str] = Field(  # type: ignore[valid-type]
        default=...,
        alias="permissions",
        description="The Privilege/Permission or list of Privileges/Permissions to grant on the given object. "
        "See https://docs.snowflake.com/en/sql-reference/sql/grant-privilege.html",
    )
    roles: Union[conlist(str, min_length=1), str] = Field(  # type: ignore[valid-type]
        default=...,
        alias="role",
        validation_alias="roles",
        description="The Role or list of Roles to grant the privileges to",
    )
    query: str = "GRANT {privileges} ON {type} {object} TO ROLE {role}"

    class Output(SnowflakeRunQueryPython.Output):
        """Output class for GrantPrivilegesOnObject"""

        query: conlist(str, min_length=1) = Field(  # type: ignore[valid-type]
            default=..., description="Query that was executed to grant privileges", validate_default=False
        )

    @model_validator(mode="before")
    def set_roles_privileges(cls, values: dict) -> dict:
        """Coerce roles and privileges to be lists if they are not already."""
        roles_value = values.get("roles") or values.get("role")
        privileges_value = values.get("privileges")

        if not (roles_value and privileges_value):
            raise ValueError("You have to specify roles AND privileges when using 'GrantPrivilegesOnObject'.")

        # coerce values to be lists
        values["roles"] = [roles_value] if isinstance(roles_value, str) else roles_value
        values["role"] = values["roles"][0]  # hack to keep the validator happy
        values["privileges"] = [privileges_value] if isinstance(privileges_value, str) else privileges_value

        return values

    @model_validator(mode="after")
    def validate_object_and_object_type(self) -> "GrantPrivilegesOnObject":
        """Validate that the object and type are set."""
        object_value = self.object
        if not object_value:
            raise ValueError("You must provide an `object`, this should be the name of the object. ")

        object_type = self.type
        if not object_type:
            raise ValueError(
                "You must provide a `type`, e.g. TABLE, VIEW, DATABASE. "
                "See https://docs.snowflake.com/en/sql-reference/sql/grant-privilege.html"
            )

        return self

    # noinspection PyMethodOverriding
    def get_query(self, role: str) -> str:
        """Build the GRANT query

        Parameters
        ----------
        role: str
            The role name

        Returns
        -------
        query : str
            The Query that performs the grant
        """
        query = self.query.format(
            privileges=",".join(self.privileges),
            type=self.type,
            object=self.object,
            role=role,
        )
        return query

    def execute(self) -> None:
        self.output.query = []
        roles = self.roles

        for role in roles:
            query = self.get_query(role)
            self.output.query.append(query)

            # Create a new instance of SnowflakeRunQueryPython with the current query
            instance = SnowflakeRunQueryPython.from_step(self, query=query)
            instance.execute()
            print(f"{instance.output = }")
            self.output.results.extend(instance.output.results)


class GrantPrivilegesOnFullyQualifiedObject(GrantPrivilegesOnObject):
    """Grant Snowflake privileges to a set of roles on a fully qualified object, i.e. `database.schema.object_name`

    This class is a subclass of `GrantPrivilegesOnObject` and is used to grant privileges on a fully qualified object.
    The advantage of using this class is that it sets the object name to be fully qualified, i.e.
    `database.schema.object_name`.

    Meaning, you can set the `database`, `schema` and `object` separately and the object name will be set to be fully
    qualified, i.e. `database.schema.object_name`.

    Example
    -------
    ```python
    GrantPrivilegesOnFullyQualifiedObject(
        database="MY_DB",
        schema="MY_SCHEMA",
        warehouse="MY_WH",
        ...
        object="MY_TABLE",
        type="TABLE",
        ...
    )
    ```

    In this example, the object name will be set to be fully qualified, i.e. `MY_DB.MY_SCHEMA.MY_TABLE`.
    If you were to use `GrantPrivilegesOnObject` instead, you would have to set the object name to be fully qualified
    yourself.
    """

    @model_validator(mode="after")
    def set_object_name(self) -> "GrantPrivilegesOnFullyQualifiedObject":
        """Set the object name to be fully qualified, i.e. database.schema.object_name"""
        # database, schema, obj_name
        db = self.database
        schema = self.model_dump()["sfSchema"]  # since "schema" is a reserved name
        obj_name = self.object

        self.object = f"{db}.{schema}.{obj_name}"

        return self


class GrantPrivilegesOnTable(GrantPrivilegesOnFullyQualifiedObject):
    """Grant Snowflake privileges to a set of roles on a table"""

    type: str = "TABLE"
    object: str = Field(
        default=...,
        alias="table",
        description="The name of the Table to grant Privileges on. This should be just the name of the table; so "
        "without Database and Schema, use sfDatabase/database and sfSchema/schema to set those instead.",
    )


class GrantPrivilegesOnView(GrantPrivilegesOnFullyQualifiedObject):
    """Grant Snowflake privileges to a set of roles on a view"""

    type: str = "VIEW"
    object: str = Field(
        default=...,
        alias="view",
        description="The name of the View to grant Privileges on. This should be just the name of the view; so "
        "without Database and Schema, use sfDatabase/database and sfSchema/schema to set those instead.",
    )



# flake8: noqa: F811
from copy import deepcopy
import os
from unittest import mock

import pytest

from pydantic import ValidationError

from koheesio.integrations.snowflake import (
    GrantPrivilegesOnObject,
    GrantPrivilegesOnTable,
    GrantPrivilegesOnView,
    SnowflakeBaseModel,
    SnowflakeRunQueryPython,
    SnowflakeStep,
    SnowflakeTableStep,
    safe_import_snowflake_connector,
)
from koheesio.integrations.snowflake.test_utils import mock_query

mock_query = mock_query

COMMON_OPTIONS = {
    "url": "hostname.com",
    "user": "user",
    "password": "password",
    "database": "db",
    "schema": "schema",
    "role": "role",
    "warehouse": "warehouse",
}


class TestGrantPrivilegesOnObject:
    options = dict(
        **COMMON_OPTIONS,
        account="42",
        object="foo",
        type="TABLE",
        privileges=["DELETE", "SELECT"],
        roles=["role_1", "role_2"],
    )

    def test_execute(self, mock_query):
        """Test that the query is correctly generated"""
        # Arrange
        del self.options["role"]  # role is not required for this test as we are setting "roles"
        mock_query.expected_data = [None]
        expected_query = [
            "GRANT DELETE,SELECT ON TABLE foo TO ROLE role_1",
            "GRANT DELETE,SELECT ON TABLE foo TO ROLE role_2",
        ]

        # Act
        kls = GrantPrivilegesOnObject(**self.options)
        output = kls.execute()

        # Assert - 2 queries are expected, result should be None
        assert output.query == expected_query
        assert output.results == [None, None]


class TestGrantPrivilegesOnTable:
    options = {**COMMON_OPTIONS, **dict(account="42", table="foo", privileges=["SELECT"], roles=["role_1"])}

    def test_execute(self, mock_query):
        """Test that the query is correctly generated"""
        # Arrange
        del self.options["role"]  # role is not required for this test as we are setting "roles"
        mock_query.expected_data = [None]
        expected_query = ["GRANT SELECT ON TABLE db.schema.foo TO ROLE role_1"]

        # Act
        kls = GrantPrivilegesOnTable(**self.options)
        output = kls.execute()

        # Assert - 1 query is expected, result should be None
        assert output.query == expected_query
        assert output.results == mock_query.expected_data


class TestGrantPrivilegesOnView:
    options = {**COMMON_OPTIONS, **dict(account="42", view="foo", privileges=["SELECT"], roles=["role_1"])}

    def test_execute(self, mock_query):
        """Test that the query is correctly generated"""
        # Arrange
        del self.options["role"]  # role is not required for this test as we are setting "roles"
        mock_query.expected_data = [None]
        expected_query = ["GRANT SELECT ON VIEW db.schema.foo TO ROLE role_1"]

        # Act
        kls = GrantPrivilegesOnView(**self.options)
        output = kls.execute()

        # Assert - 1 query is expected, result should be None
        assert output.query == expected_query
        assert output.results == mock_query.expected_data


class TestSnowflakeRunQueryPython:
    def test_mandatory_fields(self):
        """Test that query and account fields are mandatory"""
        with pytest.raises(ValidationError):
            _1 = SnowflakeRunQueryPython(**COMMON_OPTIONS)

        # sql/query and account should work without raising an error
        _2 = SnowflakeRunQueryPython(**COMMON_OPTIONS, sql="SELECT foo", account="42")
        _3 = SnowflakeRunQueryPython(**COMMON_OPTIONS, query="SELECT foo", account="42")

    def test_get_options(self):
        """Test that the options are correctly generated"""
        # Arrange
        expected_query = "SELECT foo"
        kls = SnowflakeRunQueryPython(**COMMON_OPTIONS, sql=expected_query, account="42")

        # Act
        actual_options = kls.get_options()
        query_in_options = kls.get_options(include={"query"}, by_alias=True)

        # Assert
        expected_options = {
            "account": "42",
            "database": "db",
            "password": "password",
            "role": "role",
            "schema": "schema",
            "url": "hostname.com",
            "user": "user",
            "warehouse": "warehouse",
        }
        assert actual_options == expected_options
        assert query_in_options["query"] == expected_query, "query should be returned regardless of the input"

    def test_account_populated_from_url(self):
        kls = SnowflakeRunQueryPython(**COMMON_OPTIONS, sql="SELECT * FROM table")
        assert kls.account == "hostname"

    def test_account_populated_from_url2(self):
        common_options = deepcopy(COMMON_OPTIONS)
        common_options["url"] = "https://host2.host1.snowflakecomputing.com"
        kls = SnowflakeRunQueryPython(**common_options, sql="SELECT * FROM table")
        assert kls.account == "host2"

    def test_account_populated_from_sf_url(self):
        common_options = deepcopy(COMMON_OPTIONS)
        common_options["sfURL"] = common_options.pop("url")
        kls = SnowflakeRunQueryPython(**common_options, sql="SELECT * FROM table")
        assert kls.account == "hostname"

    def test_execute(self, mock_query):
        # Arrange
        query = "SELECT * FROM two_row_table"
        expected_data = [("row1",), ("row2",)]
        mock_query.expected_data = expected_data

        # Act
        instance = SnowflakeRunQueryPython(**COMMON_OPTIONS, query=query, account="42")
        instance.execute()

        # Assert
        mock_query.assert_called_with(query)
        assert instance.output.results == expected_data

    def test_with_missing_dependencies(self):
        """Missing dependency should throw a warning first, and raise an error if execution is attempted"""
        # Arrange -- remove the snowflake connector
        with mock.patch.dict("sys.modules", {"snowflake": None}):
            from koheesio.integrations.snowflake import safe_import_snowflake_connector

            # Act & Assert -- first test for the warning, then test for the error
            match_text = "You need to have the `snowflake-connector-python` package installed"
            with pytest.warns(UserWarning, match=match_text):
                safe_import_snowflake_connector()
            with pytest.warns(UserWarning, match=match_text):
                instance = SnowflakeRunQueryPython(**COMMON_OPTIONS, query="<REDACTED>", account="42")
            with pytest.raises(RuntimeError):
                instance.execute()


class TestSnowflakeBaseModel:
    def test_get_options_using_alias(self):
        """Test that the options are correctly generated using alias"""
        k = SnowflakeBaseModel(
            sfURL="hostname.com",
            sfUser="user",
            sfPassword="password",
            sfDatabase="database",
            sfRole="role",
            sfWarehouse="warehouse",
            schema="schema",
        )
        options = k.get_options()  # alias should be used by default
        assert options["sfURL"] == "hostname.com"
        assert options["sfUser"] == "user"
        assert options["sfDatabase"] == "database"
        assert options["sfRole"] == "role"
        assert options["sfWarehouse"] == "warehouse"
        assert options["sfSchema"] == "schema"

    def test_get_options(self):
        """Test that the options are correctly generated not using alias"""
        k = SnowflakeBaseModel(
            url="hostname.com",
            user="user",
            password="password",
            database="database",
            role="role",
            warehouse="warehouse",
            schema="schema",
        )
        options = k.get_options(by_alias=False)
        assert options["url"] == "hostname.com"
        assert options["user"] == "user"
        assert options["database"] == "database"
        assert options["role"] == "role"
        assert options["warehouse"] == "warehouse"
        assert options["schema"] == "schema"

        # make sure none of the koheesio options are present
        assert "description" not in options
        assert "name" not in options

    def test_get_options_include(self):
        """Test that the options are correctly generated using include"""
        k = SnowflakeBaseModel(
            url="hostname.com",
            user="user",
            password="password",
            database="database",
            role="role",
            warehouse="warehouse",
            schema="schema",
            options={"foo": "bar"},
        )
        options = k.get_options(include={"url", "user", "description", "options"}, by_alias=False)

        # should be present
        assert options["url"] == "hostname.com"
        assert options["user"] == "user"
        assert "description" in options

        # options should be expanded
        assert "options" not in options
        assert options["foo"] == "bar"

        # should not be present
        assert "database" not in options
        assert "role" not in options
        assert "warehouse" not in options
        assert "schema" not in options


class TestSnowflakeStep:
    def test_initialization(self):
        """Test that the Step fields come through correctly"""
        # Arrange
        kls = SnowflakeStep(**COMMON_OPTIONS)

        # Act
        options = kls.get_options()

        # Assert
        assert kls.name == "SnowflakeStep"
        assert kls.description == "Expands the SnowflakeBaseModel so that it can be used as a Step"
        assert "name" not in options and "description" not in options, (
            "koheesio options should not be present in get_options"
        )


class TestSnowflakeTableStep:
    def test_initialization(self):
        """Test that the table is correctly set"""
        kls = SnowflakeTableStep(**COMMON_OPTIONS, table="table")
        assert kls.table == "table"


class TestSnowflakeConfigDir:
    @mock.patch("koheesio.integrations.snowflake.__check_access_snowflake_config_dir", return_value=False)
    @mock.patch("koheesio.integrations.snowflake.on_databricks", return_value=True)
    def test_initialization_on_databricks(self, mock_on_databricks, mock_check_access):
        """Test that the config dir is correctly set"""
        safe_import_snowflake_connector()
        assert os.environ["SNOWFLAKE_HOME"].startswith("/tmp/snowflake_tmp_")

    def test_initialization(self):
        origin_snowflake_home = os.environ.get("SNOWFLAKE_HOME")
        safe_import_snowflake_connector()
        assert os.environ.get("SNOWFLAKE_HOME") == origin_snowflake_home



import pytest

from koheesio.spark.readers.teradata import TeradataReader

pytestmark = pytest.mark.spark


class TestTeradataReader:
    common_options = {
        "url": "url",
        "user": "user",
        "password": "password",
        "dbtable": "table",
    }

    def test_get_options(self):
        tr = TeradataReader(**self.common_options)
        o = tr.get_options()

        assert o["driver"] == "com.teradata.jdbc.TeraDriver"
        assert o["fetchsize"] == 2000
        assert o["numPartitions"] == 10

    def test_execute(self, dummy_spark):
        """Method should be callable from parent class"""
        tr = TeradataReader(**self.common_options)
        assert tr.execute().df.count() == 3



"""
This module contains classes for writing data to a buffer before writing to the final destination.

The `BufferWriter` class is a base class for writers that write to a buffer first. It provides methods for writing,
reading, and resetting the buffer, as well as checking if the buffer is compressed and compressing the buffer.

The `PandasCsvBufferWriter` class is a subclass of `BufferWriter` that writes a Spark DataFrame to CSV file(s) using
Pandas. It is not meant to be used for writing huge amounts of data, but rather for writing smaller amounts of data
to more arbitrary file systems (e.g., SFTP).

The `PandasJsonBufferWriter` class is a subclass of `BufferWriter` that writes a Spark DataFrame to JSON file(s) using
Pandas. It is not meant to be used for writing huge amounts of data, but rather for writing smaller amounts of data
to more arbitrary file systems (e.g., SFTP).
"""

from __future__ import annotations

from typing import AnyStr, Literal, Optional
from abc import ABC
from functools import partial
import gzip
from os import linesep
from tempfile import SpooledTemporaryFile

from packaging import version

# noinspection PyProtectedMember
from pandas._typing import CompressionOptions as PandasCompressionOptions

from pydantic import InstanceOf

from pyspark import pandas

from koheesio.models import ExtraParamsMixin, Field, constr
from koheesio.spark import DataFrame
from koheesio.spark.writers import Writer


# pylint: disable=E1101
class BufferWriter(Writer, ABC):
    """Base class for writers that write to a buffer first, before writing to the final destination.

    `execute()` method should implement how the incoming DataFrame is written to the buffer object (e.g. BytesIO) in the
    output.

    The default implementation uses a `SpooledTemporaryFile` as the buffer. This is a file-like object that starts off
    stored in memory and automatically rolls over to a temporary file on disk if it exceeds a certain size. A
    `SpooledTemporaryFile` behaves similar to `BytesIO`, but with the added benefit of being able to handle larger
    amounts of data.

    This approach provides a balance between speed and memory usage, allowing for fast in-memory operations for smaller
    amounts of data while still being able to handle larger amounts of data that would not otherwise fit in memory.
    """

    class Output(Writer.Output, ABC):
        """Output class for BufferWriter"""

        buffer: InstanceOf[SpooledTemporaryFile] = Field(
            default_factory=partial(SpooledTemporaryFile, mode="w+b", max_size=0), exclude=True
        )

        def read(self) -> AnyStr:
            """Read the buffer"""
            self.rewind_buffer()
            data = self.buffer.read()
            self.rewind_buffer()
            return data

        def rewind_buffer(self):  # type: ignore
            """Rewind the buffer"""
            self.buffer.seek(0)
            return self

        def reset_buffer(self):  # type: ignore
            """Reset the buffer"""
            self.buffer.truncate(0)
            self.rewind_buffer()
            return self

        def is_compressed(self):  # type: ignore
            """Check if the buffer is compressed."""
            self.rewind_buffer()
            magic_number_present = self.buffer.read(2) == b"\x1f\x8b"
            self.rewind_buffer()
            return magic_number_present

        def compress(self):  # type: ignore
            """Compress the file_buffer in place using GZIP"""
            # check if the buffer is already compressed
            if self.is_compressed():
                self.logger.warn("Buffer is already compressed. Nothing to compress...")
                return self

            # compress the file_buffer
            file_buffer = self.buffer
            compressed = gzip.compress(file_buffer.read())

            # write the compressed content back to the buffer
            self.reset_buffer()
            self.buffer.write(compressed)

            return self  # to allow for chaining

    def write(self, df: DataFrame = None) -> Output:
        """Write the DataFrame to the buffer"""
        self.df = df or self.df
        if not self.df:
            raise RuntimeError("No valid Dataframe was passed")
        self.output.reset_buffer()
        self.execute()
        return self.output


# pylint: disable=C0301
class PandasCsvBufferWriter(BufferWriter, ExtraParamsMixin):
    """
    Write a Spark DataFrame to CSV file(s) using Pandas.

    Takes inspiration from https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_csv.html

    See also: https://spark.apache.org/docs/latest/sql-data-sources-csv.html#data-source-option

    Note
    ----
    This class is not meant to be used for writing huge amounts of data. It is meant to be used for writing smaller
    amounts of data to more arbitrary file systems (e.g. SFTP).

    Pyspark vs Pandas
    -----------------
    The following table shows the mapping between Pyspark, Pandas, and Koheesio properties. Note that the default values
    are mostly the same as Pyspark's `DataFrameWriter` implementation, with some exceptions (see below).

    This class implements the most commonly used properties. If a property is not explicitly implemented, it can be
    accessed through `params`.

    | PySpark Property | Default PySpark | Pandas Property | Default Pandas | Koheesio Property | Default Koheesio | Notes |
    |------------------|-----------------|-----------------|----------------|-------------------|------------------|-------|
    | maxRecordsPerFile| ...             | chunksize       | None           | max_records_per_file | ...           | Spark property name: spark.sql.files.maxRecordsPerFile |
    | sep              | ,               | sep             | ,              | sep               | ,                |       |
    | lineSep          | `\\n `          | line_terminator | os.linesep     | lineSep (alias=line_terminator) | \\n |      |
    | N/A              | ...             | index           | True           | index             | False            | Determines whether row labels (index) are included in the output |
    | header           | False           | header          | True           | header            | True             |       |
    | quote            | "               | quotechar       | "              | quote (alias=quotechar) | "          |       |
    | quoteAll         | False           | doublequote     | True           | quoteAll (alias=doublequote) | False |       |
    | escape           | `\\`              | escapechar      | None           | escapechar (alias=escape) | \\       |       |
    | escapeQuotes     | True            | N/A             | N/A            | N/A               | ...              | Not available in Pandas |
    | ignoreLeadingWhiteSpace | True     | N/A             | N/A            | N/A               | ...              | Not available in Pandas |
    | ignoreTrailingWhiteSpace | True    | N/A             | N/A            | N/A               | ...              | Not available in Pandas |
    | charToEscapeQuoteEscaping | escape or `\0` | N/A       | N/A            | N/A               | ...              | Not available in Pandas |
    | dateFormat       | `yyyy-MM-dd`      | N/A             | N/A            | N/A               | ...              | Pandas implements Timestamp, not Date |
    | timestampFormat  | `yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]` | date_format | N/A | timestampFormat (alias=date_format) | yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX] | Follows PySpark defaults |
    | timestampNTZFormat | `yyyy-MM-dd'T'HH:mm:ss[.SSS]` | N/A | N/A          | N/A               | ...              | Pandas implements Timestamp, see above |
    | compression      | None            | compression     | infer          | compression       | None             |       |
    | encoding         | utf-8           | encoding        | utf-8          | N/A               | ...              | Not explicitly implemented |
    | nullValue        | ""              | na_rep          | ""             | N/A               | ""               | Not explicitly implemented |
    | emptyValue       | ""              | na_rep          | ""             | N/A               | ""               | Not explicitly implemented |
    | N/A              | ...             | float_format    | N/A            | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | decimal         | N/A            | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | index_label     | None           | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | columns         | N/A            | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | mode            | N/A            | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | quoting         | N/A            | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | errors          | N/A            | N/A               | ...              | Not explicitly implemented |
    | N/A              | ...             | storage_options | N/A            | N/A               | ...              | Not explicitly implemented |

    differences with Pyspark:
    -------------------------
    - dateFormat -> Pandas implements Timestamp, not just Date. Hence, Koheesio sets the default to the python
        equivalent of PySpark's default.
    - compression -> Spark does not compress by default, hence Koheesio does not compress by default. Compression can
        be provided though.

    Parameters
    -----------
    header : bool, optional, default=True
        Whether to write the names of columns as the first line. In Pandas a list of strings can be given assumed to be
        aliases for the column names - this is not supported in this class. Instead, the column names as used in the
        dataframe are used as the header. Koheesio sets this default to True to match Pandas' default.
    sep : str, optional, default=,
        Field delimiter for the output file. Default is ','.
    quote : str, optional, default="
        String of length 1. Character used to quote fields. In PySpark, this is called 'quote', in Pandas this is
        called 'quotechar'. Default is '"'.
    quoteAll : bool, optional, default=False
        A flag indicating whether all values should always be enclosed in quotes in a field. Koheesio sets the default
        (False) to only escape values containing a quote character - this is Pyspark's default behavior. In Pandas, this
        is called 'doublequote'. Default is False.
    escape : str, optional, default=\\
        String of length 1. Character used to escape sep and quotechar when appropriate. Koheesio sets this default to
        `\\` to match Pyspark's default behavior. In Pandas, this field is called 'escapechar', and defaults to None.
        Default is '\\'.
    timestampFormat : str, optional, default=yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]
        Sets the string that indicates a date format for datetime objects. Koheesio sets this default to a close
        equivalent of Pyspark's default (excluding timezone information). In Pandas, this field is called 'date_format'.
        Note: Pandas does not support Java Timestamps, only Python Timestamps. The Pyspark default is
        `yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]` which mimics the iso8601 format (`datetime.isoformat()`). Default is
        '%Y-%m-%dT%H:%M:%S.%f'.
    lineSep : str, optional, default=\n
        String of length 1. Defines the character used as line separator that should be used for writing. Default is
        os.linesep.
    compression : Optional[Literal["infer", "gzip", "bz2", "zip", "xz", "zstd", "tar"]], optional, default=None
        A string representing the compression to use for on-the-fly compression of the output data. Note: Pandas sets
        this default to 'infer', Koheesio sets this default to 'None' leaving the data uncompressed by default. Can be
        set to one of 'infer', 'gzip', 'bz2', 'zip', 'xz', 'zstd', or 'tar'. See Pandas documentation for more details.
    """

    # pylint: enable=C0301

    header: bool = Field(
        default=True,
        description="Whether to write the names of columns as the first line. In Pandas a list of strings can be given "
        "assumed to be aliases for the column names - this is not supported in this class. Instead, the column names "
        "as used in the dataframe are used as the header. Koheesio sets this default to True to match Pandas' default.",
    )
    sep: constr(max_length=1) = Field(default=",", description="Field delimiter for the output file")
    quote: constr(max_length=1) = Field(
        default='"',
        description="String of length 1. Character used to quote fields. In PySpark, this is called 'quote', in Pandas "
        "this is called 'quotechar'.",
        alias="quotechar",
    )
    quoteAll: bool = Field(
        default=False,
        description="A flag indicating whether all values should always be enclosed in quotes in a field. Koheesio set "
        "the default (False) to only escape values containing a quote character - this is Pyspark's default behavior. "
        "In Pandas, this is called 'doublequote'.",
        alias="doublequote",
    )
    escape: constr(max_length=1) = Field(
        default="\\",
        description="String of length 1. Character used to escape sep and quotechar when appropriate. Koheesio sets "
        "this default to `\\` to match Pyspark's default behavior. In Pandas, this is called 'escapechar', and "
        "defaults to None.",
        alias="escapechar",
    )
    timestampFormat: str = Field(
        default="%Y-%m-%dT%H:%M:%S.%f",  # closely matches Pyspark (Java) `yyyy-MM-dd'T'HH:mm:ss[.SSS]`
        description="Sets the string that indicates a date format for datetime objects. Koheesio sets this default to "
        "a close equivalent of Pyspark's default (excluding timezone information). In Pandas, this field is called "
        "'date_format'. Note: Pandas does not support Java Timestamps, only Python Timestamps. "
        "The Pyspark default is `yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]` which mimics the iso8601 format "
        "(`datetime.isoformat()`).",
        alias="date_format",
    )
    lineSep: Optional[constr(max_length=1)] = Field(
        default=linesep,
        description="String of length 1. Defines the character used as line separator that should be used for writing.",
        alias="line_terminator",
    )
    compression: Optional[PandasCompressionOptions] = Field(
        default=None,
        description="A string representing the compression to use for on-the-fly compression of the output data."
        "Note: Pandas sets this default to 'infer', Koheesio sets this default to 'None' leaving the data uncompressed "
        "by default. Can be set to one of 'infer', 'gzip', 'bz2', 'zip', 'xz', 'zstd', or 'tar'. "
        "See Pandas documentation for more details.",
    )
    emptyValue: Optional[str] = Field(
        default="",
        description="The string to use for missing values. Koheesio sets this default to an empty string.",
    )

    nullValue: Optional[str] = Field(
        default="",
        description="The string to use for missing values. Koheesio sets this default to an empty string.",
    )

    # -- Pandas specific properties --
    index: bool = Field(
        default=False,
        description="Toggles whether to write row names (index). Default False in Koheesio - pandas default is True.",
    )

    class Output(BufferWriter.Output):
        """Output class for PandasCsvBufferWriter"""

        pandas_df: Optional[pandas.DataFrame] = Field(None, description="The Pandas DataFrame that was written")

    def get_options(self, options_type: str = "csv") -> dict:
        """Returns the options to pass to Pandas' to_csv() method."""
        try:
            import pandas as _pd

            pandas_version = version.parse(_pd.__version__)
        except ImportError:
            raise ImportError("Pandas is required to use this writer")

        # Use line_separator for pandas 2.0.0 and later
        line_sep_option_naming = "line_separator" if pandas_version >= version.parse("2.0.0") else "line_terminator"

        csv_options = {
            "header": self.header,
            "sep": self.sep,
            "quotechar": self.quote,
            "doublequote": self.quoteAll,
            "escapechar": self.escape,
            "na_rep": self.emptyValue or self.nullValue,
            line_sep_option_naming: self.lineSep,
            "index": self.index,
            "date_format": self.timestampFormat,
            "compression": self.compression,
            **self.params,
        }

        if options_type == "spark":
            csv_options["lineterminator"] = csv_options.pop(line_sep_option_naming)
        elif options_type == "koheesio_pandas_buffer_writer":
            csv_options["line_terminator"] = csv_options.pop(line_sep_option_naming)

        return csv_options

    def execute(self) -> BufferWriter.Output:
        """Write the DataFrame to the buffer using Pandas to_csv() method.
        Compression is handled by pandas to_csv() method.
        """
        # convert the Spark DataFrame to a Pandas DataFrame
        self.output.pandas_df = self.df.toPandas()

        # create csv file in memory
        file_buffer = self.output.buffer
        self.output.pandas_df.to_csv(file_buffer, **self.get_options(options_type="spark"))


# pylint: disable=C0301
class PandasJsonBufferWriter(BufferWriter, ExtraParamsMixin):
    """
    Write a Spark DataFrame to JSON file(s) using Pandas.

    Takes inspiration from https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_json.html

    Note
    ----
    This class is not meant to be used for writing huge amounts of data. It is meant to be used for writing
    smaller amounts of data to more arbitrary file systems (e.g. SFTP).

    Parameters
    -----------
    orient : Literal["split", "records", "index", "columns", "values", "table"]
        Format of the resulting JSON string. Default is 'records'.
    lines : bool
        Format output as one JSON object per line. Only used when orient='records'. Default is True.
        - If true, the output will be formatted as one JSON object per line.
        - If false, the output will be written as a single JSON object.
        Note: this value is only used when orient='records' and will be ignored otherwise.
    date_format : Literal["iso", "epoch"]
        Type of date conversion. Default is 'iso'. See `Date and Timestamp Formats` for a detailed description and
        more information.
    double_precision : int
        Number of decimal places for encoding floating point values. Default is 10.
    force_ascii : bool
        Force encoded string to be ASCII. Default is True.
    compression : Optional[Literal["gzip"]]
        A string representing the compression to use for on-the-fly compression of the output data.
        Koheesio sets this default to 'None' leaving the data uncompressed. Can be set to gzip' optionally.
        Other compression options are currently not supported by Koheesio for JSON output.

    Other Possible Parameters
    -------------------------

    ### Date and Timestamp Formats in JSON
    The `date_format` and `date_unit` parameters in pandas `to_json()` method are used to control the representation of
    dates and timestamps in the resulting JSON.

    - `date_format`: This parameter determines the format of date strings. It accepts two options: 'epoch' and 'iso'.
        - 'epoch': Dates are represented as the number of milliseconds since the epoch (1970-01-01).
        - 'iso': Dates are represented in ISO8601 format, 'YYYY-MM-DDTHH:MM:SS'.
        In Pandas the default value depends on the `orient` parameter. For `orient='table'`, the default is 'iso'. For
        all other `orient` values, the default is 'epoch'.
        However, in Koheesio, the default is set to 'iso' irrespective of the `orient` parameter.

    - `date_unit`: This parameter specifies the time unit for encoding timestamps and datetime objects. It accepts
        four options: 's' for seconds, 'ms' for milliseconds, 'us' for microseconds, and 'ns' for nanoseconds.
        The default is 'ms'. Note that this parameter is ignored when `date_format='iso'`.

    ### Orient Parameter
    The `orient` parameter is used to specify the format of the resulting JSON string. Each option is useful in
    different scenarios depending on whether you need to preserve the index, data types, and/or column names of the
    original DataFrame. The set of possible orients is:

    - 'records': List of dictionaries with each dictionary representing a row of data. This is the default orient in
        Koheesio.
        - Does not preserve the index.
        - If `lines=True` (default), each record is written as a separate line in the output file.
            - Example:
            ```json
            {"column1": value1, "column2": value2}
            {"column1": value3, "column2": value4}
            ```
        - If `lines=False`, all records are written within a single JSON array.
            - Example:
            ```json
            [{"column1": value1, "column2": value2}, {"column1": value3, "column2": value4}]
            ```
    - 'split': Dictionary containing indexes, columns, and data.
        - Preserves data types and indexes of the original DataFrame.
        - Example:
        ```json
        {"index": [index], "columns": [columns], "data": [values]}
        ```
    - 'index': Dictionary with DataFrame indexes as keys and dictionaries with column names and values as values.
        - Preserves the index.
        - Example:
        ```json
        {"index1": {"column1": value1, "column2": value2}}
        ```
    - 'columns': Dictionary with column names as keys and dictionaries with indexes and values as values.
        - Preserves data types and indexes of the original DataFrame.
        - Example:
        ```json
        {"column1": {"index1": value1}, "column2": {"index1": value1}}
        ```
    - 'values': Just the values in the DataFrame.
        - Does not preserve the index or columns.
        - Example:
        ```json
        [[value1, value2], ..., [valueN-1, valueN]]
        ```
    - 'table': Dictionary with 'schema' and 'data' keys. 'schema' has information about data types and index, 'data'
        has the actual data.
        - Preserves data types and indexes of the original DataFrame.
        - Example:
        ```json
        {
          "schema": {
            "fields": [
              {
                "name": "index",
                "type": "dtype"
              }
            ],
            "primaryKey": ["index"]
          },
          "pandas_version": "1.4.0",
          "data": [
            {
              "column1": "value1",
              "column2": "value2"
            }
          ]
        }
        ```

    Note
    ----
    For 'records' orient, set `lines` to True to write each record as a separate line. The pandas output will
    then match the PySpark output (orient='records' and lines=True parameters).

    References
    ----------
    - [Pandas DataFrame to_json documentation](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_json.html)
    - [Pandas IO tools (text, CSV, HDF5, …) documentation](https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html)
    """

    # pylint: enable=C0301

    orient: Literal["split", "records", "index", "columns", "values", "table"] = Field(
        default="records",
        description="Format of the resulting JSON string. Default is 'records'.",
    )
    lines: bool = Field(
        default=True,
        description="Format output as one JSON object per line. Only used when orient='records'. Default is True.",
    )
    date_format: Literal["iso", "epoch"] = Field(
        default="iso",
        description="Type of date conversion. Default is 'iso'.",
    )
    double_precision: int = Field(
        default=10,
        description="Number of decimal places for encoding floating point values. Default is 10.",
    )
    force_ascii: bool = Field(
        default=True,
        description="Force encoded string to be ASCII. Default is True.",
    )
    columns: Optional[list[str]] = Field(
        default=None,
        description="The columns to write. If None, all columns will be written.",
    )
    compression: Optional[Literal["gzip"]] = Field(
        default=None,
        description="A string representing the compression to use for on-the-fly compression of the output data."
        "Koheesio sets this default to 'None' leaving the data uncompressed by default. "
        "Can be set to 'gzip' optionally.",
    )

    class Output(BufferWriter.Output):
        """Output class for PandasJsonBufferWriter"""

        pandas_df: Optional[pandas.DataFrame] = Field(None, description="The Pandas DataFrame that was written")

    def get_options(self) -> dict:
        """Returns the options to pass to Pandas' to_json() method."""
        json_options = {
            "orient": self.orient,
            "date_format": self.date_format,
            "double_precision": self.double_precision,
            "force_ascii": self.force_ascii,
            "lines": self.lines,
            **self.params,
        }

        # ignore the 'lines' parameter if orient is not 'records'
        if self.orient != "records":
            del json_options["lines"]

        return json_options

    def execute(self) -> BufferWriter.Output:
        """Write the DataFrame to the buffer using Pandas to_json() method."""
        df = self.df
        if self.columns:
            df = df[self.columns]

        # convert the Spark DataFrame to a Pandas DataFrame
        self.output.pandas_df = df.toPandas()

        # create json file in memory
        file_buffer = self.output.buffer
        self.output.pandas_df.to_json(file_buffer, **self.get_options())

        # compress the buffer if compression is set
        if self.compression:
            self.output.compress()



import pytest

from pyspark.sql import functions as F

from koheesio.spark import AnalysisException, DataFrame
from koheesio.spark.readers.delta import DeltaTableReader

pytestmark = pytest.mark.spark


def test_delta_table_reader(spark):
    df = DeltaTableReader(table="delta_test_table").read()
    df.show()
    actual = df.head().asDict()
    expected = {"id": 0}

    assert isinstance(df, DataFrame)
    assert actual == expected


def test_delta_table_cdf_reader(spark, streaming_dummy_df, random_uuid):
    table_name = f"delta_test_table_cdf_{random_uuid}"

    spark.sql(
        f"CREATE TABLE IF NOT EXISTS {table_name} (id INT) "
        "USING DELTA "
        "TBLPROPERTIES (delta.enableChangeDataFeed = true)"
    )
    spark.sql(f"INSERT INTO {table_name} VALUES (10)")
    spark.sql(f"INSERT INTO {table_name} VALUES (0)")
    spark.sql(f"UPDATE {table_name} SET id = 100 WHERE id = 0")
    spark.sql(f"DELETE FROM {table_name} WHERE id = 10")
    spark.sql(f"UPDATE {table_name} SET id = 0 WHERE id = 100")

    cdf_reader = DeltaTableReader(table=table_name, read_change_feed=True, starting_version=0)
    df = cdf_reader.read()
    df.orderBy("_commit_timestamp").show(truncate=False)
    df.show()
    assert df.count() == 7


# FIXME: causing other tests to fail
# def test_delta_table_cdf_stream_reader(spark, checkpoint_folder, streaming_dummy_df):
#     spark.sql("INSERT INTO delta_test_table VALUES (10)")
#     spark.sql("UPDATE delta_test_table SET id = 100 WHERE id = 0")
#     spark.sql("DELETE FROM delta_test_table WHERE id = 10")
#     spark.sql("UPDATE delta_test_table SET id = 0 WHERE id = 100")
#
#     cdf_reader = DeltaTableReader(
#         table="delta_test_table",
#         read_change_feed=True,
#         ignore_changes=True,
#         streaming=True,
#         starting_version=0,
#         filter_cond=col("_change_type") != "insert",
#     )
#
#     cdf_reader.read().writeStream.format("console").start().awaitTermination(60)


def test_delta_reader_view(spark):
    reader = DeltaTableReader(table="delta_test_table")

    with pytest.raises(AnalysisException):
        _ = spark.table(reader.view)
        # In Spark remote session the above statetment will not raise an exception
        _ = spark.table(reader.view).take(1)
    reader.read()
    df = spark.table(reader.view)
    assert df.count() == 10


def test_delta_reader_failed():
    table_name = "delta_test_table"
    with pytest.raises(ValueError):
        DeltaTableReader(table=table_name, starting_version=3, starting_timestamp="2023-10-02").read()

    with pytest.raises(ValueError):
        DeltaTableReader(table=table_name, read_change_feed=True).read()


@pytest.fixture(scope="session")
def select_and_filter_table(spark):
    df = spark.createDataFrame(
        [["Michael", "SG", 6], ["Scottie", "SF", 6], ["Steve", "SG", 5], ["Tony", "F", 3], ["Dennis", "C", 5]],
        schema=["name", "role", "rings"],
    )

    df.write.format("delta").saveAsTable("select_and_filter_test")


def test_string_filter_cond(spark, select_and_filter_table):
    expected = {"name": "Michael", "rings": 6, "role": "SG"}
    reader = DeltaTableReader(table="select_and_filter_test", filter_cond="name = 'Michael'")

    actual = reader.read().head().asDict()
    assert actual == expected


@pytest.mark.parametrize(
    "filter_cond,expected",
    [
        ("""F.col("rings").eqNullSafe(3)""", {"name": "Tony", "rings": 3, "role": "F"}),
        ("""(F.col("rings") > 3) & (F.col("name").startswith("St"))""", {"name": "Steve", "rings": 5, "role": "SG"}),
    ],
)
def test_filter_cond(spark, select_and_filter_table, filter_cond, expected):
    filter_cond = eval(filter_cond)
    reader = DeltaTableReader(table="select_and_filter_test", filter_cond=filter_cond)
    actual = reader.read().head().asDict()
    assert actual == expected
    assert reader.df.select(F.col("name")).head().asDict() == {"name": expected["name"]}


@pytest.mark.parametrize(
    "columns,expected_schema",
    [
        (["name", "rings"], {"name", "rings"}),
        (["role", "rings"], {"role", "rings"}),
        (None, {"name", "rings", "role"}),
        ("name", {"name"}),
    ],
)
def test_select_columns(spark, select_and_filter_table, columns, expected_schema):
    reader = DeltaTableReader(table="select_and_filter_test", columns=columns)
    actual_schema = set(reader.read().columns)
    assert actual_schema == expected_schema



"""
This module contains several HTTP Steps that can be used to perform API Calls to HTTP endpoints

Example
-------
```python
from koheesio.steps.http import HttpGetStep

response = (
    HttpGetStep(url="https://google.com").execute().json_payload
)
```

In the above example, the `response` variable will contain the JSON response from the HTTP request.
"""

from typing import Any, Dict, Generator, List, Optional, Union
import contextlib
from enum import Enum
import json

import requests  # type: ignore[import-untyped]
from urllib3.util import Retry

from koheesio import Step
from koheesio.models import (
    ExtraParamsMixin,
    Field,
    SecretStr,
    field_serializer,
    field_validator,
    model_validator,
)

__all__ = [
    "HttpMethod",
    "HttpStep",
    "HttpGetStep",
    "HttpPostStep",
    "HttpPutStep",
    "HttpDeleteStep",
    "PaginatedHttpGetStep",
]


class HttpMethod(str, Enum):
    """
    Enumeration of allowed http methods
    """

    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"

    @classmethod
    def from_string(cls, value: str) -> str:
        """Allows for getting the right Method Enum by simply passing a string value
        This method is not case-sensitive
        """
        return getattr(cls, value.upper())


class HttpStep(Step, ExtraParamsMixin):
    """
    Can be used to perform API Calls to HTTP endpoints

    Authorization
    -------------
    The optional `auth_header` parameter in HttpStep allows you to pass an authorization header, such as a bearer token.
    For example: `auth_header = "Bearer <token>"`.

    The `auth_header` value is stored as a `SecretStr` object to prevent sensitive information from being displayed in logs.

    Of course, authorization can also just be passed as part of the regular `headers` parameter.

    For example, either one of these parameters would semantically be the same:
    ```python
    headers = {
        "Authorization": "Bearer <token>",
        "Content-Type": "application/json",
    }
    ```
    # or
    auth_header = "Bearer <token>"
    ```

    The `auth_header` parameter is useful when you want to keep the authorization separate from the other headers, for
    example when your implementation requires you to pass some custom headers in addition to the authorization header.

    > Note: The `auth_header` parameter can accept any authorization header value, including basic authentication
        tokens, digest authentication strings, NTLM, etc.

    Understanding Retries
    ----------------------
    This class includes a built-in retry mechanism for handling temporary issues, such as network errors or server
    downtime, that might cause the HTTP request to fail. The retry mechanism is controlled by two parameters:
    `max_retries` and `backoff_factor`, and will only be triggered for error codes 502,503 and 504.

    - `max_retries` determines the number of retries after the initial request. For example, if `max_retries` is set to
        4, the request will be attempted a total of 5 times (1 initial attempt + 4 retries). If `max_retries` is set to
        0, no retries will be attempted, and the request will be tried only once.

    - `backoff_factor` controls the rate at which the delay increases for each subsequent retry. If `backoff_factor` is
        set to 2 (the default), the delay will double with each retry. If `backoff` is set to 1, the delay between
        retries will remain constant. Changing the `backoff_factor` value affects how quickly the delay increases.

    Given the default values of `max_retries=3` and `backoff=2`, the delays between retries would be 2 seconds,
    4 seconds, and 8 seconds, respectively. This results in a total delay of 14 seconds before all retries are
    exhausted.

    Parameters
    ----------
    url : str, required
        API endpoint URL.
    headers : Dict[str, Union[str, SecretStr]], optional, default={"Content-Type": "application/json"}
        Request headers.
    auth_header : Optional[SecretStr], optional, default=None
        Authorization header. An optional parameter that can be used to pass an authorization, such as a bearer token.
    data : Union[Dict[str, str], str], optional, default={}
        Data to be sent along with the request.
    timeout : int, optional, default=3
        Request timeout. Defaults to 3 seconds.
    method : Union[str, HttpMethod], required, default='get'
        What type of Http call to perform. One of 'get', 'post', 'put', 'delete'. Defaults to 'get'.
    session : requests.Session, optional, default=requests.Session()
        Existing requests session object to be used for making HTTP requests. If not provided, a new session object
        will be created.
    params : Optional[Dict[str, Any]]
        Set of extra parameters that should be passed to the HTTP request. Note: any kwargs passed to the class will be
        added to this dictionary.
    max_retries : int, optional, default=3
        Maximum number of retries before giving up. Defaults to 3.
    backoff_factor : float, optional, default=2
        Backoff factor for retries. Defaults to 2.

    Output
    ------
    response_raw : Optional[requests.Response]
        The raw requests.Response object returned by the appropriate requests.request() call.
    response_json : Optional[Union[Dict, List]]
        The JSON response for the request.
    raw_payload : Optional[str]
        The raw response for the request.
    status_code : Optional[int]
        The status return code of the request.
    """

    url: str = Field(
        default=...,
        description="API endpoint URL",
        alias="uri",
    )
    headers: Dict[str, Union[str, SecretStr]] = Field(
        default={"Content-Type": "application/json"},
        description="Request headers",
        alias="header",
    )
    auth_header: Optional[SecretStr] = Field(
        default=None,
        description="[Optional] Authorization header",
        alias="authorization_header",
        examples=["Bearer <token>"],
    )
    data: Union[Dict[str, str], str] = Field(
        default_factory=dict, description="[Optional] Data to be sent along with the request", alias="body"
    )
    params: Optional[Dict[str, Any]] = Field(  # type: ignore[assignment]
        default_factory=dict,
        description="[Optional] Set of extra parameters that should be passed to HTTP request",
    )
    timeout: int = Field(default=3, description="[Optional] Request timeout")
    method: Union[str, HttpMethod] = Field(
        default=HttpMethod.GET,
        description="What type of Http call to perform. One of 'get', 'post', 'put', 'delete'. Defaults to 'get'.",
    )
    session: requests.Session = Field(
        default_factory=requests.Session,
        description=(
            "Existing requests session object to be used for making HTTP requests. If not provided, a new session "
            "object will be created."
        ),
        exclude=True,
        repr=False,
    )
    max_retries: int = Field(
        default=3,
        description="Maximum number of retries before giving up. Defaults to 3.",
    )
    backoff_factor: float = Field(
        default=2,
        description="Backoff factor for retries. Defaults to 2.",
    )

    class Output(Step.Output):
        """Output class for HttpStep"""

        response_raw: Optional[requests.Response] = Field(
            default=None,
            alias="response",
            description="The raw requests.Response object returned by the appropriate requests.request() call",
        )
        response_json: Optional[Union[Dict, List]] = Field(
            default=None, alias="json_payload", description="The JSON response for the request"
        )
        raw_payload: Optional[str] = Field(
            default=None, alias="response_text", description="The raw response for the request"
        )
        status_code: Optional[int] = Field(default=None, description="The status return code of the request")

        @property
        def json_payload(self) -> Union[dict, list, None]:
            """Alias for response_json"""
            return self.response_json

    @field_validator("method")
    def get_proper_http_method_from_str_value(cls, method_value: str) -> str:
        """Converts string value to HttpMethod enum value"""
        if isinstance(method_value, str):
            try:
                method_value = HttpMethod.from_string(method_value)
            except AttributeError as e:
                raise AttributeError(
                    "Only values from HttpMethod class are allowed! "
                    f"Provided value: '{method_value}', allowed values: {', '.join(HttpMethod.__members__.keys())}"
                ) from e

        return method_value

    @model_validator(mode="after")
    def encode_sensitive_headers(self) -> "HttpStep":
        """
        Encode potentially sensitive data into pydantic.SecretStr class to prevent them
        being displayed as plain text in logs.
        """
        if auth_header := self.auth_header:
            # ensure the token is preceded with the word 'Bearer'
            self.headers["Authorization"] = auth_header
            del self.auth_header
        if auth := self.headers.get("Authorization"):
            self.headers["Authorization"] = auth if isinstance(auth, SecretStr) else SecretStr(auth)
        return self

    @field_serializer("headers", when_used="json")
    def decode_sensitive_headers(self, headers: dict) -> dict:
        """
        Authorization headers are being converted into SecretStr under the hood to avoid dumping any
        sensitive content into logs by the `encode_sensitive_headers` method.

        However, when calling the `get_headers` method, the SecretStr should be converted back to
        string, otherwise sensitive info would have looked like '**********'.

        This method decodes values of the `headers` dictionary that are of type SecretStr into plain text.
        """
        for k, v in headers.items():
            headers[k] = v.get_secret_value() if isinstance(v, SecretStr) else v
        return headers

    def get_headers(self) -> dict:
        """
        Dump headers into JSON without SecretStr masking.
        """
        return json.loads(self.model_dump_json()).get("headers")

    def set_outputs(self, response: requests.Response) -> None:
        """
        Types of response output
        """
        self.output.response_raw = response
        self.output.raw_payload = response.text
        self.output.status_code = response.status_code

        # Only decode non empty payloads to avoid triggering decoding error unnecessarily.
        if self.output.raw_payload:
            try:
                self.output.response_json = response.json()

            except json.decoder.JSONDecodeError as e:
                self.log.error(f"An error occurred while processing the JSON payload. Error message:\n{e.msg}")

    def get_options(self) -> dict:
        """options to be passed to requests.request()"""
        return {
            "url": self.url,
            "headers": self.get_headers(),
            "data": self.data,
            "timeout": self.timeout,
            **self.params,  # type: ignore
        }

    def _configure_session(self) -> None:
        """Configure session with current retry settings"""
        retries = Retry(
            total=self.max_retries,
            connect=None,  # Only retry on status codes
            read=None,  # Only retry on status codes
            redirect=0,  # No redirect retries
            status=self.max_retries,
            backoff_factor=self.backoff_factor,
            status_forcelist=[502, 503, 504],
            allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE"]),
            respect_retry_after_header=True,
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retries)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

    @contextlib.contextmanager
    def _request(
        self, method: Optional[HttpMethod] = None, stream: bool = False
    ) -> Generator[requests.Response, None, None]:
        """
        Executes the HTTP request with retry logic.

        Actual http_method execution is abstracted into this method.
        This is to avoid unnecessary code duplication. Allows to centrally log, set outputs, and validated.

        This method will try to execute `requests.request` up to `self.max_retries` times. If `self.request()` raises
        an exception with error code 502,503 or 504, it logs a warning message and the error message, then waits for
        `(self.backoff_factor ** i)` seconds before retrying, where the delay increases exponentially after each failed
        attempt.

        If `self.request()` still fails after `self.max_retries` attempts, it logs an error message and re-raises the
        last exception that was caught.

        This is a good way to handle temporary issues that might cause `self.request()` to fail, such as network errors
        or server downtime. The exponential backoff ensures that you're not constantly bombarding a server with
        requests if it's struggling to respond.

        Parameters
        ----------
        method : HttpMethod
            Optional parameter that allows calls to different HTTP methods and bypassing class level `method`
            parameter.
        stream : bool
            Whether to stream the response content. Defaults to False.

        Raises
        ------
        requests.RequestException, requests.HTTPError
            The last exception that was caught if `requests.request()` fails after `self.max_retries` attempts.
        """
        _method = (method or self.method).value.upper()
        self.log.debug(f"Making {_method} request to {self.url} with headers {self.headers}")
        options = self.get_options()

        self._configure_session()

        with self.session.request(method=_method, **options, stream=stream) as response:
            response.raise_for_status()
            self.log.debug(f"Received response with status code {response.status_code} and body {response.text}")

            try:
                yield response
            finally:
                self.log.debug("Request context manager exiting")

    # noinspection PyMethodOverriding
    def get(self) -> requests.Response:
        """Execute an HTTP GET call"""
        self.method = HttpMethod.GET
        with self.request() as response:
            return response

    def post(self) -> requests.Response:
        """Execute an HTTP POST call"""
        self.method = HttpMethod.POST
        with self.request() as response:
            return response

    def put(self) -> requests.Response:
        """Execute an HTTP PUT call"""
        self.method = HttpMethod.PUT
        with self.request() as response:
            return response

    def delete(self) -> requests.Response:
        """Execute an HTTP DELETE call"""
        self.method = HttpMethod.DELETE
        with self.request() as response:
            return response

    def execute(self) -> None:
        """
        Executes the HTTP request.

        This method simply calls `self.request()`, which includes the retry logic. If `self.request()` raises an
        exception, it will be propagated to the caller of this method.

        Raises
        ------
        requests.RequestException, requests.HTTPError
            The last exception that was caught if `self.request()` fails after `self.max_retries` attempts.
        """
        with self._request() as response:
            self.log.info(f"HTTP request to {self.url}, status code {response.status_code}")
            self.set_outputs(response)


class HttpGetStep(HttpStep):
    """send GET requests

    Example
    -------
    ```python
    response = (
        HttpGetStep(url="https://google.com").execute().json_payload
    )
    ```
    In the above example, the `response` variable will contain the JSON response from the HTTP request.
    """

    method: HttpMethod = HttpMethod.GET


class HttpPostStep(HttpStep):
    """send POST requests"""

    method: HttpMethod = HttpMethod.POST


class HttpPutStep(HttpStep):
    """send PUT requests"""

    method: HttpMethod = HttpMethod.PUT


class HttpDeleteStep(HttpStep):
    """send DELETE requests"""

    method: HttpMethod = HttpMethod.DELETE


class PaginatedHttpGetStep(HttpGetStep):
    """
    Represents a paginated HTTP GET step.

    Parameters
    ----------
    paginate : bool, optional
        Whether to paginate the API response. Defaults to False.
    pages : int, optional
        Number of pages to paginate. Defaults to 1.
    offset : int, optional
        Offset for paginated API calls. Offset determines the starting page. Defaults to 1.
    limit : int, optional
        Limit for paginated API calls. Defaults to 100.
    """

    paginate: Optional[bool] = Field(
        default=False,
        description="Whether to paginate the API response. Defaults to False. When set to True, the API response will "
        "be paginated. The url should contain a named 'page' parameter for example: "
        "api.example.com/data?page={page}",
    )
    pages: Optional[int] = Field(default=1, description="Number of pages to paginate. Defaults to 1")
    offset: Optional[int] = Field(
        default=1,
        description="Offset for paginated API calls. Offset determines the starting page. Defaults to 1. "
        "The url can (optionally) contain a named 'offset' parameter, for example: "
        "api.example.com/data?offset={offset}",
    )
    limit: Optional[int] = Field(
        default=100,
        description="Limit for paginated API calls. The url should (optionally) contain a named limit parameter, "
        "for example: api.example.com/data?limit={limit}",
    )

    def _adjust_params(self) -> Dict[str, Any]:
        """
        Adjusts the parameters by removing the 'paginate' key.

        Returns
        -------
        dict
            The adjusted parameters.
        """
        return {k: v for k, v in self.params.items() if k not in ["paginate"]}  # type: ignore

    def get_options(self) -> dict:
        """
        Returns the options to be passed to the requests.request() function.

        Returns
        -------
        dict
            The options.
        """
        options = {
            "url": self.url,
            "headers": self.get_headers(),
            "data": self.data,
            "timeout": self.timeout,
            **self._adjust_params(),  # type: ignore
        }

        return options

    def _url(self, basic_url: str, page: Optional[int] = None) -> str:
        """
        Adds additional parameters to the URL.

        Parameters
        ----------
        basic_url : str
            The basic URL without any parameters.
        page : int, optional
            The page number for pagination. Defaults to None.

        Returns
        -------
        str
            The URL with additional parameters.
        """
        url_params = self._adjust_params()

        if self.paginate:
            url_params.update(
                {
                    "offset": self.offset,
                    "limit": self.limit,
                    "page": page,
                }
            )

        return basic_url.format(**url_params)

    def execute(self) -> None:
        """
        Executes the HTTP GET request and handles pagination.

        Returns
        -------
        HttpGetStep.Output
            The output of the HTTP GET request.
        """
        # Set up pagination parameters
        offset, pages = (self.offset, self.pages + 1) if self.paginate else (1, 1)  # type: ignore
        data = []
        _basic_url = self.url

        for page in range(offset, pages):  # type: ignore[arg-type]
            if self.paginate:
                self.log.info(f"Fetching page {page} of {pages - 1}")

            self.url = self._url(basic_url=_basic_url, page=page)

            with self._request() as response:
                if isinstance(response_json := response.json(), list):
                    data += response_json
                else:
                    data.append(response_json)

        self.url = _basic_url
        self.output.response_json = data
        self.output.response_raw = None
        self.output.raw_payload = None
        self.output.status_code = None



"""
Utility functions
"""

from typing import Any, Callable, Dict, Optional, Tuple
import datetime
from functools import partial, wraps
from importlib import import_module
import inspect
from pathlib import Path
from sys import version_info as PYTHON_VERSION
import uuid
import warnings

__all__ = [
    "get_args_for_func",
    "get_project_root",
    "import_class",
    "get_random_string",
    "convert_str_to_bool",
]


PYTHON_MINOR_VERSION = PYTHON_VERSION.major + PYTHON_VERSION.minor / 10
"""float: Python minor version as a float (e.g. 3.7)"""


def get_args_for_func(func: Callable, params: Dict) -> Tuple[Callable, Dict[str, Any]]:
    """Helper function that matches keyword arguments (params) on a given function

    This function uses inspect to extract the signature on the passed Callable, and then uses `functools.partial` to
     construct a new Callable (partial) function on which the input was mapped.

    Example
    -------
    ```python
    input_dict = {"a": "foo", "b": "bar"}


    def example_func(a: str):
        return a


    func, kwargs = get_args_for_func(example_func, input_dict)
    ```

    In this example,
    - `func` would be a callable with the input mapped toward it (i.e. can be called like any normal function)
    - `kwargs` would be a dict holding just the output needed to be able to run the function (e.g. {"a": "foo"})

    Parameters
    ----------
    func: Callable
        The function to inspect
    params: Dict
        Dictionary with keyword values that will be mapped on the 'func'

    Returns
    -------
    Tuple[Callable, Dict[str, Any]]
        - Callable
            a partial() func with the found keyword values mapped toward it
        - Dict[str, Any]
            the keyword args that match the func
    """
    _kwargs = {k: v for k, v in params.items() if k in inspect.getfullargspec(func).args}
    return (
        partial(func, **_kwargs),
        _kwargs,
    )


def get_project_root() -> Path:
    """Returns project root path."""
    cmd = Path(__file__)
    return Path([i for i in cmd.parents if i.as_uri().endswith("src")][0]).parent


def import_class(module_class: str) -> Any:
    """Import class and module based on provided string.

    Parameters
    ----------
    module_class module+class to be imported.

    Returns
    -------
    object  Class from specified input string.

    """
    module_path, class_name = module_class.rsplit(".", 1)
    module = import_module(module_path)

    return getattr(module, class_name)


def get_random_string(length: int = 64, prefix: Optional[str] = None) -> str:
    """Generate a random string of specified length"""
    if prefix:
        return f"{prefix}_{uuid.uuid4().hex}"[0:length]
    return f"{uuid.uuid4().hex}"[0:length]


def convert_str_to_bool(value: str) -> Any:
    """Converts a string to a boolean if the string is either 'true' or 'false'"""
    if isinstance(value, str) and (v := value.lower()) in ["true", "false"]:
        value = v == "true"
    return value


def utc_now() -> datetime.datetime:
    """Get current time in UTC"""
    if PYTHON_MINOR_VERSION < 3.11:
        return datetime.datetime.utcnow()
    return datetime.datetime.now(datetime.timezone.utc)



"""
Test the Substring columns transformation
"""

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.substring import Substring
from koheesio.spark.utils import show_string

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)

data_with_strings = [[1, 2, "hello"], [3, 4, "world"], [3, 4, None]]
schema_with_strings = ["c1", "c2", "c3"]


@pytest.mark.parametrize(
    "input_values,data,schema,expected",
    [
        (
            # case 1 : column specified -> trim a specific column
            dict(column="c3", start=1, length=3, target_column="result"),
            # input data and schema
            data_with_strings,
            schema_with_strings,
            # expected output
            ["hel", "wor", None],
        ),
        (
            # case 2 : column specified -> trim multiple columns (no strings)
            # all numeric columns will be converted to strings, c3 is ignored
            dict(columns=["c1", "c2"], start=1, length=1),
            # input data and schema
            data_with_strings,
            schema_with_strings,
            # expected output
            [
                {"c1": "1", "c2": "2", "c3": "hello"},
                {"c1": "3", "c2": "4", "c3": "world"},
                {"c1": "3", "c2": "4", "c3": None},
            ],
        ),
        (
            # case 3 : example from class docstring
            dict(column="c1", start=3, length=4, target_column="result"),
            # input data and schema
            [["skyscraper"]],
            ["c1"],
            # expected output
            ["yscr"],
        ),
        (
            # case 4 : no length specified
            dict(column="c3", start=3, target_column="result"),
            # input data and schema
            data_with_strings,
            schema_with_strings,
            # expected output
            ["llo", "rld", None],
        ),
    ],
)
def test_substring(input_values, data, schema, expected, spark):
    input_df = spark.createDataFrame(data, schema)

    substring = Substring(**input_values)
    output_df = substring.transform(input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{show_string(output_df, 20, 20, False)}")

    if target_column := substring.target_column:
        actual = [row.asDict()[target_column] for row in output_df.collect()]
    else:
        actual = [row.asDict() for row in output_df.collect()]

    assert actual == expected



"""
This module provides functionality to download files from URLs specified in a Spark DataFrame column and store the
downloaded file paths in a new column. It leverages the `DownloadFileStep` class to handle the file download process
and supports various write modes to manage existing files.

Classes
-------
DownloadFileFromUrlTransformation
    A transformation class that downloads content from URLs in the specified column
    and stores the downloaded file paths in a new column.

Write Modes
-----------
`DownloadFileFromUrlTransformation` supports the following write modes:

=== "OVERWRITE"

    ```python
    DownloadFileFromUrlTransformation(
        ...
        mode=FileWriteMode.OVERWRITE,
    )
    ```

    - If the file exists, it will be overwritten.
    - If it does not exist, a new file will be created.

    <small>(this is the default mode)</small>

=== "APPEND"

    ```python
    DownloadFileFromUrlTransformation(
        ...
        mode=FileWriteMode.APPEND,
    )
    ```

    - If the file exists, the new data will be appended to it.
    - If it does not exist, a new file will be created.

    <br>

=== "IGNORE"

    ```python
    DownloadFileFromUrlTransformation(
        ...
        mode=FileWriteMode.IGNORE,
    )
    ```

    - If the file exists, the method will return without writing anything.
    - If it does not exist, a new file will be created.

    <br>

=== "EXCLUSIVE"

    ```python
    DownloadFileFromUrlTransformation(
        ...
        mode=FileWriteMode.EXCLUSIVE,
    )
    ```

    - If the file exists, an error will be raised.
    - If it does not exist, a new file will be created.

    <br>

=== "BACKUP"

    ```python
    DownloadFileFromUrlTransformation(
        ...
        mode=FileWriteMode.BACKUP,
    )
    ```

    - If the file exists, a backup will be created and the original file will be overwritten.
    - If it does not exist, a new file will be created.

    <br>
"""

from typing import Union

from koheesio.models import DirectoryPath, Field
from koheesio.spark import Column
from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils.common import get_column_name
from koheesio.steps.download_file import DownloadFileStep, FileWriteMode


class DownloadFileFromUrlTransformation(ColumnsTransformationWithTarget):
    """
    Downloads content from URLs in the specified column and stores the downloaded file paths in a new column.

    Example
    -------
    Example usage of the `DownloadFileFromUrlTransformation` class:

    ```python
    from pyspark.sql import SparkSession
    from koheesio.spark.transformations.download_files import (
        DownloadFileFromUrlTransformation,
    )
    from koheesio.steps.download_file import FileWriteMode

    spark = SparkSession.builder.appName(
        "DownloadFilesExample"
    ).getOrCreate()
    df = spark.createDataFrame(
        [
            ("http://example.com/file1.txt",),
            ("http://example.com/file2.txt",),
        ],
        ["url"],
    )

    transformation = DownloadFileFromUrlTransformation(
        column="url",
        target_column="downloaded_file_path",
        mode=FileWriteMode.OVERWRITE,
        download_path="/path/to/download",
    )

    transformed_df = transformation.transform(df)
    transformed_df.show()
    ```

    In this example, the `DownloadFileFromUrlTransformation` class is used to download files from the URLs specified in
    the `url` column of the DataFrame `df`. The downloaded file paths are stored in a new column named
    `downloaded_file_path`. The downloaded files are saved to the `/path/to/download` directory with the `OVERWRITE`
    write mode. (The `OVERWRITE` mode is the default mode.)

    ### Input DataFrame:

    | url                          |
    |------------------------------|
    | http://example.com/file1.txt |
    | http://example.com/file2.txt |

    ### Output DataFrame:

    | url                          | downloaded_file_path |
    |------------------------------|----------------------|
    | http://example.com/file1.txt | download/file1.txt   |
    | http://example.com/file2.txt | download/file2.txt   |

    Since the `DownloadFileFromUrlTransformation` class is a `ColumnsTransformationWithTarget`, the `transform` method
    is used to apply the transformation to the input DataFrame `df`. Alternatively, the `execute` method can be used
    to apply the transformation in place, or the class can be used on a `df.transform()` call.

    I.e.:
    ```python
    # Using the transform method
    transformed_df = transformation.transform(df)
    # Using the execute method
    transformed_df = (
        transformation.execute().df
    )  # note: while doing this, df needs to be passed in the constructor
    # Using the df.transform() method
    transformed_df = df.transform(transformation)
    ```

    Parameters
    ----------
    column : Union[Column, str]
        The column that holds the URLs to download.
    download_path : str
        The local directory path where the file will be downloaded to.
    chunk_size : int, optional, default=8192
        The size (in bytes) of the chunks to download the file in, must be greater than 16.
    mode : FileWriteMode, optional, default=FileWriteMode.OVERWRITE
        Write mode: overwrite, append, ignore, exclusive, or backup.

    Write Modes
    -----------
    The `DownloadFileFromUrlTransformation` supports the following write modes:

    - OVERWRITE
    - APPEND
    - IGNORE
    - EXCLUSIVE
    - BACKUP
    """

    column: Union[Column, str] = Field(
        default="",
        description="The column that holds the URLs to download.",
    )
    download_path: DirectoryPath = Field(
        ..., description="The local directory path where the file will be downloaded to."
    )
    chunk_size: int = Field(
        8192,
        ge=16,
        description="The size (in bytes) of the chunks to download the file in, must be greater than or equal to 16.",
    )
    mode: FileWriteMode = Field(
        default=FileWriteMode.OVERWRITE,
        description="Write mode: overwrite, append, ignore, exclusive, or backup.",
    )

    class Output(ColumnsTransformationWithTarget.Output):
        download_file_paths: dict[str, str] = Field(
            default_factory=dict, description="The downloaded file paths per URL."
        )

    def func(self, partition: set[str]) -> None:
        """
        Takes a set of urls and downloads the files from the URLs in the specified column.

        Parameters
        ----------
        partition : set[str]
            A set of URLs to download the files from.
        """
        for url in partition:
            step = DownloadFileStep(
                url=url, download_path=self.download_path, mode=self.mode, chunk_size=self.chunk_size
            )
            step.execute()
            self.output.download_file_paths[url] = step.output.download_file_path.relative_to(
                self.download_path.parent
            ).as_posix()

    def execute(self) -> Output:
        """
        Download files from URLs in the specified column.
        """
        # Collect the URLs from the DataFrame and process them
        source_column_name = self.column
        if not isinstance(source_column_name, str):
            source_column_name = get_column_name(source_column_name)

        partition = {row.asDict()[source_column_name] for row in self.df.select(self.column).collect()}  # type: ignore
        self.func(partition)

        # Using join, re-add the download_file_paths to the DataFrame in the target column
        url_df = self.spark.createDataFrame(
            data=self.output.download_file_paths.items(), schema=f"{self.column} string, {self.target_column} string"
        )
        self.output.df = (
            self.df.join(
                other=url_df,  # type: ignore
                on=source_column_name,
                how="left",
            ).select(*self.df.columns, self.target_column)  # type: ignore
        )



# noinspection PyUnresolvedReferences
"""
Transformations to cast a column or set of columns to a given datatype.

Each one of these have been vetted to throw warnings when wrong datatypes are passed (to prevent errors in any job or
pipeline).

Furthermore, detailed tests have been added to ensure that types are actually compatible as prescribed.

Concept
-------
* One can use the CastToDataType class directly, or use one of the more specific subclasses.
* Each subclass is compatible with 'run_for_all' - meaning that if no column or columns are specified, it will be run
    for all compatible data types.
* Compatible Data types are specified in the ColumnConfig class of each subclass, and are documented in the docstring
    of each subclass.

See class docstrings for more information

Note
----
Dates, Arrays and Maps are not supported by this module.

- for dates, use the [koheesio.spark.transformations.date_time](date_time/index.md) module
- for arrays, use the [koheesio.spark.transformations.arrays](arrays.md) module

Classes
-------
CastToDatatype
    Cast a column or set of columns to a given datatype
CastToByte
    Cast to Byte (a.k.a. tinyint)
CastToShort
    Cast to Short (a.k.a. smallint)
CastToInteger
    Cast to Integer (a.k.a. int)
CastToLong
    Cast to Long (a.k.a. bigint)
CastToFloat
    Cast to Float (a.k.a. real)
CastToDouble
    Cast to Double
CastToDecimal
    Cast to Decimal (a.k.a. decimal, numeric, dec, BigDecimal)
CastToString
    Cast to String
CastToBinary
    Cast to Binary (a.k.a. byte array)
CastToBoolean
    Cast to Boolean
CastToTimestamp
    Cast to Timestamp

Note
----
The following parameters are common to all classes in this module:

Parameters
----------
columns : ListOfColumns
    Name of the source column(s). Alias: column
target_column : str
    Name of the target column or alias if more than one column is specified. Alias: target_alias
datatype : str or SparkDatatype
    Datatype to cast to. Choose from SparkDatatype Enum (only needs to be specified in CastToDatatype, all other
    classes have a fixed datatype)
"""

from typing import Union

from pyspark.sql import Column

from koheesio.models import Field, conint, field_validator, model_validator
from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils import SparkDatatype


class CastToDatatype(ColumnsTransformationWithTarget):
    """
    Cast a column or set of columns to a given datatype

    Wrapper around pyspark.sql.Column.cast

    Concept
    -------
    This class acts as the base class for all the other CastTo* classes. It is compatible with 'run_for_all' - meaning
    that if no column or columns are specified, it will be run for all compatible data types.

    Example
    -------
    __input_df__:

    | c1 | c2 |
    |----|----|
    | 1  | 2  |
    | 3  | 4  |

    ```python
    output_df = CastToDatatype(
        column="c1",
        datatype="string",
        target_alias="c1",
    ).transform(input_df)
    ```

    __output_df__:

    | c1   | c2 |
    |------|----|
    | "1"  | 2  |
    | "3"  | 4  |

    In the example above, the column `c1` is cast to a string datatype. The column `c2` is not affected.

    Parameters
    ----------
    columns : ListOfColumns
        Name of the source column(s). Alias: column
    datatype : str or SparkDatatype
        Datatype to cast to. Choose from SparkDatatype Enum
    target_column : str
        Name of the target column or alias if more than one column is specified. Alias: target_alias
    """

    datatype: Union[str, SparkDatatype] = Field(default=..., description="Datatype. Choose from SparkDatatype Enum")

    @field_validator("datatype")
    def validate_datatype(cls, datatype_value: Union[str, SparkDatatype]) -> SparkDatatype:  # type: ignore
        """Validate the datatype."""
        # handle string input
        try:
            if isinstance(datatype_value, str):
                datatype_value: SparkDatatype = SparkDatatype.from_string(datatype_value)
                return datatype_value

            # and let SparkDatatype handle the rest
            datatype_value: SparkDatatype = SparkDatatype.from_string(datatype_value.value)

        except AttributeError as e:
            raise AttributeError(f"Invalid datatype: {datatype_value}") from e

        return datatype_value

    def func(self, column: Column) -> Column:
        # This is to let the IDE explicitly know that the datatype is not a string, but a `SparkDatatype` Enum
        datatype: SparkDatatype = self.datatype  # type: ignore
        return column.cast(datatype.spark_type())


class CastToByte(CastToDatatype):
    """
    Cast to Byte (a.k.a. tinyint)

    Represents 1-byte signed integer numbers. The range of numbers is from -128 to 127.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * boolean
    * timestamp
    * decimal
    * double
    * float
    * long
    * integer
    * short

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * timestamp
        range of values too small for timestamp to have any meaning
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToByte class."""

        run_for_all_data_type = [
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.TIMESTAMP,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.BYTE


class CastToShort(CastToDatatype):
    """
    Cast to Short (a.k.a. smallint)

    Represents 2-byte signed integer numbers. The range of numbers is from -32768 to 32767.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * integer
    * long
    * float
    * double
    * decimal
    * string
    * boolean
    * timestamp
    * date
    * void

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * timestamp
        range of values too small for timestamp to have any meaning
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToShort class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.TIMESTAMP,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.SHORT


class CastToInteger(CastToDatatype):
    """
    Cast to Integer (a.k.a. int)

    Represents 4-byte signed integer numbers. The range of numbers is from -2147483648 to 2147483647.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * long
    * float
    * double
    * decimal
    * boolean
    * timestamp

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToInteger class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
            SparkDatatype.TIMESTAMP,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.INTEGER


class CastToLong(CastToDatatype):
    """
    Cast to Long (a.k.a. bigint)

    Represents 8-byte signed integer numbers. The range of numbers is from -9223372036854775808 to 9223372036854775807.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * long
    * float
    * double
    * decimal
    * boolean
    * timestamp

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToLong class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
            SparkDatatype.TIMESTAMP,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.LONG


class CastToFloat(CastToDatatype):
    """
    Cast to Float (a.k.a. real)

    Represents 4-byte single-precision floating point numbers. The range of numbers is from -3.402823E38 to 3.402823E38.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * double
    * decimal
    * boolean

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * timestamp
        precision is lost (use CastToDouble instead)
    * string
        converts to null
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToFloat class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.FLOAT


class CastToDouble(CastToDatatype):
    """
    Cast to Double

    Represents 8-byte double-precision floating point numbers. The range of numbers is from -1.7976931348623157E308 to
    1.7976931348623157E308.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * float
    * decimal
    * boolean
    * timestamp

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToDouble class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
            SparkDatatype.TIMESTAMP,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.DOUBLE


class CastToDecimal(CastToDatatype):
    """
    Cast to Decimal (a.k.a. decimal, numeric, dec, BigDecimal)

    Represents arbitrary-precision signed decimal numbers. Backed internally by `java.math.BigDecimal`. A BigDecimal
    consists of an arbitrary precision integer unscaled value and a 32-bit integer scale.

    The DecimalType must have fixed precision (the maximum total number of digits) and scale (the number of digits on
    the right of dot). For example, (5, 2) can support the value from [-999.99 to 999.99].

    The precision can be up to 38, the scale must be less or equal to precision.

    Spark sets the default precision and scale to (10, 0) when creating a DecimalType. However, when inferring schema
    from decimal.Decimal objects, it will be DecimalType(38, 18).

    For compatibility reasons, Koheesio sets the default precision and scale to (38, 18) when creating a DecimalType.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * float
    * double
    * boolean
    * timestamp
    * date
    * string
    * void
    * decimal
        spark will convert existing decimals to null if the precision and scale doesn't fit the data

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * date
        converts to null
    * void
        skipped by default

    Parameters
    ----------
    columns : ListOfColumns, optional, default=*
        Name of the source column(s). Alias: column
    target_column : str
        Name of the target column or alias if more than one column is specified. Alias: target_alias
    precision : conint(gt=0, le=38), optional, default=38
        the maximum (i.e. total) number of digits (default: 38). Must be > 0.
    scale : conint(ge=0, le=18), optional, default=18
        the number of digits on right side of dot. (default: 18). Must be >= 0.
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToDecimal class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BOOLEAN,
            SparkDatatype.TIMESTAMP,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.DECIMAL
    precision: conint(gt=0, le=38) = Field(
        default=38,
        description="The maximum total number of digits (precision) of the decimal. Must be > 0. Default is 38",
    )
    scale: conint(ge=0, le=18) = Field(
        default=18,
        description="The number of digits to the right of the decimal point (scale). Must be >= 0. Default is 18",
    )

    @model_validator(mode="after")
    def validate_scale_and_precisions(self) -> "CastToDecimal":
        """Validate the precision and scale values."""
        precision_value = self.precision
        scale_value = self.scale

        if scale_value == precision_value:
            self.log.warning("scale and precision are equal, this will result in a null value")
        if scale_value > precision_value:
            raise ValueError("scale must be < precision")

        return self

    def func(self, column: Column) -> Column:
        return column.cast(self.datatype.spark_type(precision=self.precision, scale=self.scale))


class CastToString(CastToDatatype):
    """
    Cast to String

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * float
    * double
    * decimal
    * binary
    * boolean
    * timestamp
    * date
    * array
    * map
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToString class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.BINARY,
            SparkDatatype.BOOLEAN,
            SparkDatatype.TIMESTAMP,
            SparkDatatype.DATE,
            SparkDatatype.ARRAY,
            SparkDatatype.MAP,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.STRING


class CastToBinary(CastToDatatype):
    """
    Cast to Binary (a.k.a. byte array)

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * float
    * double
    * decimal
    * boolean
    * timestamp
    * date
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * string

    Koheesio will skip the transformation for these unless specific columns of these types are given as input:

    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToBinary class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.STRING,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.BINARY


class CastToBoolean(CastToDatatype):
    """
    Cast to Boolean

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * float
    * double
    * decimal
    * timestamp

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * string
        converts to null
    * date
        converts to null
    * void
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToBoolean class."""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.TIMESTAMP,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.STRING,
            SparkDatatype.DATE,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.BOOLEAN


class CastToTimestamp(CastToDatatype):
    """
    Cast to Timestamp

    Numeric time stamp is the number of seconds since January 1, 1970 00:00:00.000000000 UTC.
    Not advised to be used on small integers, as the range of values is too small for timestamp to have any meaning.

    For more fine-grained control over the timestamp format, use the `date_time` module. This allows for parsing strings
    to timestamps and vice versa.

    See Also
    --------
    * [koheesio.spark.transformations.date_time](date_time/index.md)
    * https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html#timestamp-pattern

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    * binary
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * integer
    * long
    * float
    * double
    * decimal
    * date

    Spark doesn't error on these, but will cast to null or otherwise give mangled data. Hence, Koheesio will skip the
    transformation for these unless specific columns of these types are given as input:

    * boolean:
        range of values too small for timestamp to have any meaning
    * byte:
        range of values too small for timestamp to have any meaning
    * string:
        converts to null in most cases, use `date_time` module instead
    * short:
        range of values too small for timestamp to have any meaning
    * void:
        skipped by default
    """

    class ColumnConfig(CastToDatatype.ColumnConfig):
        """Set the data types that are compatible with the CastToTimestamp class."""

        run_for_all_data_type = [
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.DATE,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.BOOLEAN,
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.STRING,
            SparkDatatype.VOID,
        ]

    datatype: Union[str, SparkDatatype] = SparkDatatype.TIMESTAMP
